# Copyright 2014-2017 Red Hat, Inc. and others
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Contains a view that allows API users to search packages."""

from pyramid.view import view_config

from bodhi.server import log, buildsys


def get_all_packages():
    """
    Return a list of all packages in Koji.

    Returns:
        list: The list of package_names from the koji.listPackages() call.
    """
    log.debug('Fetching list of all packages...')
    koji = buildsys.get_session()
    return [pkg['package_name'] for pkg in koji.listPackages()]


@view_config(route_name='search_packages', renderer='json',
             request_method='GET')
def search_packages(request):
    """
    Search for packages that match the given term GET query parameter.

    Deprecated: This view is unused by the web UI and the CLI and might be a good candidate for
    removal in a future major release.

    Args:
        request (pyramid.request): The current web request.
    Returns:
        list: A list of dictionaries with keys 'id', 'label', and 'value' all indexing koji
            listPackages() results that match the search term.
    """
    packages = get_all_packages()
    return [{'id': p, 'label': p, 'value': p} for p in packages
            if request.GET['term'] in p]



# -*- coding: utf-8 -*-
# Copyright © 2013-2017 Red Hat, Inc. and others.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Bodhi's configuration and configuration loading and validation mechanisms."""
from datetime import datetime
import os
import logging
import binascii

from pyramid import settings
from pyramid.paster import get_appsettings
import cryptography.fernet
import six


log = logging.getLogger('bodhi')


def get_configfile():
    """
    Return a path to a config file, if found.

    Return the path to a config file, with a heirarchy of preferential paths. It searches first
    for development.ini if found. if not found, it will return /etc/bodhi/production.ini if it
    exists. Otherwise, it returns None.

    Returns:
        basestring or None: The path of a config file, or None if no config file is found.
    """
    configfile = None
    setupdir = os.path.join(os.path.dirname(os.path.dirname(__file__)), '..')
    for cfg in (os.path.join(setupdir, 'development.ini'),
                '/etc/bodhi/production.ini'):
        if os.path.exists(cfg):
            configfile = cfg
            break
    else:
        log.error("Unable to find configuration to load!")
    return configfile


def _generate_list_validator(splitter=' ', validator=six.text_type):
    """Return a function that takes a value and interprets it to be a list with the given splitter.

    This function generates a function that can take a string and interpret it as a list by
    splitting with the given splitter string. Each element of the resulting list is then validated
    with the given validator.

    Args:
        splitter (basestring): A string to use to split the input into a list.
        validator (function):  A function to apply to each element of the list to validate it.
    Returns:
        function: A validator function that accepts an argument to be validated.
    """
    def _validate_list(value):
        """Validate that the value is a list or can be split into a list, and validate its elements.

        This function will validate that the given value is a list, or it will use the splitter to
        turn it into a list. Once it is a list, it will use the validator on each element of the
        list.

        Args:
            value (basestring or list): The list to be validated.
        Returns:
            unicode: The interpreted list.
        Raises:
            ValueError: If validator fails on any of the list's elements.
        """
        if isinstance(value, six.string_types):
            value = [idx.strip() for idx in value.split(splitter) if idx.strip()]

        if not isinstance(value, list):
            raise ValueError('"{}" cannot be intepreted as a list.'.format(value))

        # Run the validator on each element of the list.
        value = [validator(v) for v in value]

        return value

    return _validate_list


def _validate_bool(value):
    """Return a bool version of value.

    This function will ensure that value is a bool, or that it is a string that can be interpreted
    as a bool. It will return a bool. If it cannot do that, it will raise ValueError.

    Args:
        value (basestring or bool): The value to be validated as a bool.
    Returns:
        bool: The boolean interpretation of value.
    Raises:
        ValueError: If value cannot be interpreted as a boolean.
    """
    if isinstance(value, six.string_types):
        # Recent versions of Pyramid define a settings.falsey, but version 1.5.6 does not so its
        # values have been backported here for the False case. Pyramid defines an asbool(), but it
        # will not raise any error for strings that aren't in the truthy or falsey lists, and we
        # want strict validation.
        if value.lower().strip() in settings.truthy:
            return True
        elif value.lower().strip() in ('f', 'false', 'n', 'no', 'off', '0'):
            return False
        else:
            raise ValueError('"{}" cannot be interpreted as a boolean value.'.format(value))

    if not isinstance(value, bool):
        raise ValueError('"{}" is not a bool or a string.'.format(value))

    return value


def _validate_color(value):
    """Ensure that value is a valid expression of a color, in the form #dddddd.

    Return the value if it is a valid color expression, or raise ValueError.

    Args:
        value (basestring): The color to be validated.
    Returns:
        unicode: The color.
    Raises:
        ValueError: If value is not in the form #dddddd.
    """
    e = ValueError('"{}" is not a valid color expression.'.format(value))

    if not isinstance(value, six.string_types):
        raise e
    if not len(value) == 7:
        raise e
    if value[0] != '#':
        raise e
    try:
        int(value[-6:], 16)
    except ValueError:
        raise e

    return six.text_type(value)


def _validate_fernet_key(value):
    """Ensure the value is not CHANGEME, that it is a Fernet key, and convert it to a str.

    This function is used to ensure that secret values in the config have been set by the user to
    something other than the default of CHANGEME and that the value can be used as a Fernet key. It
    is converted to str before returning.

    Args:
        value (basestring): The value to be validated.
    Returns:
        str: The value.
    Raises:
        ValueError: If value is "CHANGEME" or if it cannot be used as a Fernet key.
    """
    _validate_secret(value)

    if isinstance(value, six.text_type):
        value = value.encode('utf-8')

    try:
        engine = cryptography.fernet.Fernet(value)
        # This will raise a ValueError if value is not suitable as a Fernet key.
        engine.encrypt(b'a secret test string')
    except (TypeError, binascii.Error):
        raise ValueError('Fernet key must be 32 url-safe base64-encoded bytes.')

    return value


def _validate_none_or(validator):
    """Return a function that will ensure a value is None or passes validator.

    This function returns a function that will take a single argument, value, and will ensure
    that value is None or that it passes the given validator.

    Args:
        validator (function): A function to apply when value is not None.
    Returns:
        function: A validator function that accepts an argument to be validated.
    """
    def _validate(value):
        if value is None:
            return value

        return validator(value)

    return _validate


def validate_path(value):
    """Ensure that value is an existing path on the local filesystem and return it.

    Use os.path.exists to ensure that value is an existing path. Return the value if it is, else
    raise ValueError.

    Args:
        value (basestring): The path to be validated.
    Returns:
        unicode: The path.
    Raises:
        ValueError: If os.path.exists returns False.
    """
    if not os.path.exists(value):
        raise ValueError('"{}" does not exist.'.format(value))

    return six.text_type(value)


def _validate_rstripped_str(value):
    """
    Ensure that value is a str that is rstripped of the / character.

    Args:
        value (six.text_type): The value to be validated and rstripped.
    Returns:
        six.text_type: The rstripped value.
    """
    value = six.text_type(value)
    return value.rstrip('/')


def _validate_secret(value):
    """Ensure that the value is not CHANGEME and convert it to unicode.

    This function is used to ensure that secret values in the config have been set by the user to
    something other than the default of CHANGEME.

    Args:
        value (basestring): The value to be validated.
    Returns:
        unicode: The value.
    Raises:
        ValueError: If value is "CHANGEME".
    """
    if value == 'CHANGEME':
        raise ValueError('This setting must be changed from its default value.')

    return six.text_type(value)


def _validate_tls_url(value):
    """Ensure that the value is a string that starts with https://.

    Args:
        value (basestring): The value to be validated.
    Returns:
        unicode: The value.
    Raises:
        ValueError: If value is not a string starting with https://.
    """
    if not isinstance(value, six.string_types) or not value.startswith('https://'):
        raise ValueError('This setting must be a URL starting with https://.')

    return six.text_type(value)


class BodhiConfig(dict):
    """
    A dictionary interface to the Bodhi configuration.

    This class defines defaults for most of Bodhi's settings, and also provides validation that
    converts them to the expected types.
    """

    loaded = False

    _defaults = {
        'acl_system': {
            'value': 'dummy',
            'validator': six.text_type},
        'acl_dummy_committer': {
            'value': None,
            'validator': _validate_none_or(six.text_type)},
        'admin_groups': {
            'value': ['proventesters', 'security_respons', 'bodhiadmin', 'sysadmin-main'],
            'validator': _generate_list_validator()},
        'admin_packager_groups': {
            'value': ['provenpackager', 'releng', 'security_respons'],
            'validator': _generate_list_validator()},
        'authtkt.secret': {
            'value': 'CHANGEME',
            'validator': _validate_secret},
        'authtkt.secure': {
            'value': True,
            'validator': _validate_bool},
        'authtkt.timeout': {
            'value': 86400,
            'validator': int},
        'badge_ids': {
            'value': [],
            'validator': _generate_list_validator('|')},
        'base_address': {
            'value': 'https://admin.fedoraproject.org/updates/',
            'validator': six.text_type},
        'bodhi_email': {
            'value': 'updates@fedoraproject.org',
            'validator': six.text_type},
        'bodhi_password': {
            'value': None,
            'validator': _validate_none_or(six.text_type)},
        'buglink': {
            'value': 'https://bugzilla.redhat.com/show_bug.cgi?id=%s',
            'validator': six.text_type},
        'bugtracker': {
            'value': None,
            'validator': _validate_none_or(six.text_type)},
        'bugzilla_api_key': {
            'value': None,
            'validator': _validate_none_or(six.text_type)},
        'buildroot_limit': {
            'value': 31,
            'validator': int},
        'buildsystem': {
            'value': 'dev',
            'validator': six.text_type},
        'bz_products': {
            'value': [],
            'validator': _generate_list_validator(',')},
        'bz_server': {
            'value': 'https://bugzilla.redhat.com/xmlrpc.cgi',
            'validator': six.text_type},
        'cache_dir': {
            'value': None,
            'validator': _validate_none_or(validate_path)},
        'captcha.background_color': {
            'value': '#ffffff',
            'validator': _validate_color},
        'captcha.font_color': {
            'value': '#000000',
            'validator': _validate_color},
        'captcha.font_path': {
            'value': '/usr/share/fonts/liberation/LiberationMono-Regular.ttf',
            'validator': validate_path},
        'captcha.font_size': {
            'value': 36,
            'validator': int},
        'captcha.image_height': {
            'value': 80,
            'validator': int},
        'captcha.image_width': {
            'value': 300,
            'validator': int},
        'captcha.padding': {
            'value': 5,
            'validator': int},
        'captcha.secret': {
            'value': None,
            'validator': _validate_none_or(_validate_fernet_key)},
        'captcha.ttl': {
            'value': 300,
            'validator': int},
        'clean_old_composes': {
            'value': True,
            'validator': _validate_bool},
        'container.destination_registry': {
            'value': 'registry.fedoraproject.org',
            'validator': six.text_type},
        'container.source_registry': {
            'value': 'candidate-registry.fedoraproject.org',
            'validator': six.text_type},
        'cors_connect_src': {
            'value': 'https://*.fedoraproject.org/ wss://hub.fedoraproject.org:9939/',
            'validator': six.text_type},
        'cors_origins_ro': {
            'value': '*',
            'validator': six.text_type},
        'cors_origins_rw': {
            'value': 'https://bodhi.fedoraproject.org',
            'validator': six.text_type},
        'critpath_pkgs': {
            'value': [],
            'validator': _generate_list_validator()},
        'critpath.min_karma': {
            'value': 2,
            'validator': int},
        'critpath.num_admin_approvals': {
            'value': 2,
            'validator': int},
        'critpath.stable_after_days_without_negative_karma': {
            'value': 14,
            'validator': int},
        'critpath.type': {
            'value': None,
            'validator': _validate_none_or(six.text_type)},
        'datagrepper_url': {
            'value': 'https://apps.fedoraproject.org/datagrepper',
            'validator': six.text_type},
        'default_email_domain': {
            'value': 'fedoraproject.org',
            'validator': six.text_type},
        'disable_automatic_push_to_stable': {
            'value': (
                'Bodhi is disabling automatic push to stable due to negative karma. The '
                'maintainer may push manually if they determine that the issue is not severe.'),
            'validator': six.text_type},
        'dogpile.cache.arguments.filename': {
            'value': '/var/cache/bodhi-dogpile-cache.dbm',
            'validator': six.text_type},
        'dogpile.cache.backend': {
            'value': 'dogpile.cache.dbm',
            'validator': six.text_type},
        'dogpile.cache.expiration_time': {
            'value': 100,
            'validator': int},
        'exclude_mail': {
            'value': ['autoqa', 'taskotron'],
            'validator': _generate_list_validator()},
        'fedmenu.data_url': {
            'value': 'https://apps.fedoraproject.org/js/data.js',
            'validator': six.text_type},
        'fedmenu.url': {
            'value': 'https://apps.fedoraproject.org/fedmenu',
            'validator': six.text_type},
        'fedmsg_enabled': {
            'value': False,
            'validator': _validate_bool},
        'file_url': {
            'value': 'https://download.fedoraproject.org/pub/fedora/linux/updates',
            'validator': six.text_type},
        'fmn_url': {
            'value': 'https://apps.fedoraproject.org/notifications/',
            'validator': six.text_type},
        'important_groups': {
            'value': ['proventesters', 'provenpackager,' 'releng', 'security_respons', 'packager',
                      'bodhiadmin'],
            'validator': _generate_list_validator()},
        'initial_bug_msg': {
            'value': '%s has been submitted as an update to %s. %s',
            'validator': six.text_type},
        'greenwave_api_url': {
            'value': 'https://greenwave-web-greenwave.app.os.fedoraproject.org/api/v1.0',
            'validator': _validate_rstripped_str},
        'waiverdb_api_url': {
            'value': 'https://waiverdb-web-waiverdb.app.os.fedoraproject.org/api/v1.0',
            'validator': _validate_rstripped_str},
        'waiverdb.access_token': {
            'value': None,
            'validator': _validate_none_or(six.text_type)},
        'koji_web_url': {
            'value': 'https://koji.fedoraproject.org/koji/',
            'validator': _validate_tls_url},
        'koji_hub': {
            'value': 'https://koji.stg.fedoraproject.org/kojihub',
            'validator': str},
        'krb_ccache': {
            'value': None,
            'validator': _validate_none_or(str)},
        'krb_keytab': {
            'value': None,
            'validator': _validate_none_or(str)},
        'krb_principal': {
            'value': None,
            'validator': _validate_none_or(str)},
        'legal_link': {
            'value': '',
            'validator': six.text_type},
        'libravatar_dns': {
            'value': False,
            'validator': _validate_bool},
        'libravatar_enabled': {
            'value': True,
            'validator': _validate_bool},
        'mail.templates_basepath': {
            'value': 'bodhi:server/email/templates/',
            'validator': six.text_type},
        'mako.directories': {
            'value': 'bodhi:server/templates',
            'validator': six.text_type},
        'mandatory_packager_groups': {
            'value': ['packager'],
            'validator': _generate_list_validator()},
        'mash_dir': {
            'value': None,
            'validator': _validate_none_or(six.text_type)},
        'mash_stage_dir': {
            'value': None,
            'validator': _validate_none_or(six.text_type)},
        'max_concurrent_mashes': {
            'value': 2,
            'validator': int},
        'max_update_length_for_ui': {
            'value': 30,
            'validator': int},
        'message_id_email_domain': {
            'value': 'admin.fedoraproject.org',
            'validator': six.text_type},
        'not_yet_tested_epel_msg': {
            'value': (
                'This update has not yet met the minimum testing requirements defined in the '
                '<a href="https://fedoraproject.org/wiki/EPEL_Updates_Policy">EPEL Update Policy'
                '</a>'),
            'validator': six.text_type},
        'not_yet_tested_msg': {
            'value': (
                'This update has not yet met the minimum testing requirements defined in the '
                '<a href="https://fedoraproject.org/wiki/Package_update_acceptance_criteria">'
                'Package Update Acceptance Criteria</a>'),
            'validator': six.text_type},
        'openid.provider': {
            'value': 'https://id.fedoraproject.org/openid/',
            'validator': six.text_type},
        'openid.sreg_required': {
            'value': 'email',
            'validator': six.text_type},
        'openid.success_callback': {
            'value': 'bodhi.server.security:remember_me',
            'validator': six.text_type},
        'openid.url': {
            'value': 'https://id.fedoraproject.org/',
            'validator': six.text_type},
        'openid_template': {
            'value': '{username}.id.fedoraproject.org',
            'validator': six.text_type},
        'pagure_url': {
            'value': 'https://src.fedoraproject.org/pagure/',
            'validator': _validate_tls_url},
        'pdc_url': {
            'value': 'https://pdc.fedoraproject.org/',
            'validator': _validate_tls_url},
        'pkgdb_url': {
            'value': 'https://admin.fedoraproject.org/pkgdb',
            'validator': six.text_type},
        'prefer_ssl': {
            'value': None,
            'validator': _validate_none_or(bool)},
        'privacy_link': {
            'value': '',
            'validator': six.text_type},
        'pungi.basepath': {
            'value': '/etc/bodhi',
            'validator': six.text_type},
        'pungi.cmd': {
            'value': '/usr/bin/pungi-koji',
            'validator': six.text_type},
        'pungi.conf.module': {
            'value': 'pungi.module.conf',
            'validator': six.text_type},
        'pungi.conf.rpm': {
            'value': 'pungi.rpm.conf',
            'validator': six.text_type},
        'pungi.extracmdline': {
            'value': [],
            'validator': _generate_list_validator()},
        'pungi.labeltype': {
            'value': 'Update',
            'validator': six.text_type},
        'query_wiki_test_cases': {
            'value': False,
            'validator': _validate_bool},
        'release_team_address': {
            'value': 'bodhiadmin-members@fedoraproject.org',
            'validator': six.text_type},
        'resultsdb_api_url': {
            'value': 'https://taskotron.fedoraproject.org/resultsdb_api/',
            'validator': six.text_type},
        'session.secret': {
            'value': 'CHANGEME',
            'validator': _validate_secret},
        'site_requirements': {
            'value': 'dist.rpmdeplint dist.upgradepath',
            'validator': six.text_type},
        'skopeo.cmd': {
            'value': '/usr/bin/skopeo',
            'validator': six.text_type,
        },
        'skopeo.extra_copy_flags': {
            'value': '',
            'validator': six.text_type,
        },
        'smtp_server': {
            'value': None,
            'validator': _validate_none_or(six.text_type)},
        'sqlalchemy.url': {
            'value': 'sqlite:////var/cache/bodhi.db',
            'validator': six.text_type},
        'stable_bug_msg': {
            'value': ('%s has been pushed to the %s repository. If problems still persist, please '
                      'make note of it in this bug report.'),
            'validator': six.text_type},
        'stable_from_batched_msg': {
            'value': ('This update has been dequeued from batched and is now entering stable.'),
            'validator': six.text_type},
        'stacks_enabled': {
            'value': False,
            'validator': _validate_bool},
        'stats_blacklist': {
            'value': ['bodhi', 'anonymous', 'autoqa', 'taskotron'],
            'validator': _generate_list_validator()},
        'system_users': {
            'value': ['bodhi', 'autoqa', 'taskotron'],
            'validator': _generate_list_validator()},
        'test_case_base_url': {
            'value': 'https://fedoraproject.org/wiki/',
            'validator': six.text_type},
        'testing_approval_msg_based_on_karma': {
            'value': ('This update has reached the stable karma threshold and can be pushed to '
                      'stable now if the maintainer wishes.'),
            'validator': six.text_type
        },
        'testing_approval_msg': {
            'value': ('This update has reached %d days in testing and can be pushed to stable now '
                      'if the maintainer wishes'),
            'validator': six.text_type},
        'testing_bug_epel_msg': {
            'value': (
                '\nSee https://fedoraproject.org/wiki/QA:Updates_Testing for\ninstructions on how '
                'to install test updates.\nYou can provide feedback for this update here: %s'),
            'validator': six.text_type},
        'testing_bug_msg': {
            'value': (
                '\nSee https://fedoraproject.org/wiki/QA:Updates_Testing for\ninstructions on how '
                'to install test updates.\nYou can provide feedback for this update here: %s'),
            'validator': six.text_type},
        'top_testers_timeframe': {
            'value': 7,
            'validator': int},
        'test_gating.required': {
            'value': False,
            'validator': _validate_bool},
        'test_gating.url': {
            'value': '',
            'validator': six.text_type},
        'updateinfo_rights': {
            'value': 'Copyright (C) {} Red Hat, Inc. and others.'.format(datetime.now().year),
            'validator': six.text_type},
        'wait_for_repo_sig': {
            'value': False,
            'validator': _validate_bool},
        'wiki_url': {
            'value': 'https://fedoraproject.org/w/api.php',
            'validator': six.text_type},
    }

    def __getitem__(self, *args, **kw):
        """Ensure the config is loaded, and then call the superclass __getitem__."""
        if not self.loaded:
            self.load_config()
        return super(BodhiConfig, self).__getitem__(*args, **kw)

    def get(self, *args, **kw):
        """Ensure the config is loaded, and then call the superclass get."""
        if not self.loaded:
            self.load_config()
        return super(BodhiConfig, self).get(*args, **kw)

    def pop(self, *args, **kw):
        """Ensure the config is loaded, and then call the superclass pop."""
        if not self.loaded:
            self.load_config()
        return super(BodhiConfig, self).pop(*args, **kw)

    def copy(self, *args, **kw):
        """Ensure the config is loaded, and then call the superclass copy."""
        if not self.loaded:
            self.load_config()
        return super(BodhiConfig, self).copy(*args, **kw)

    def load_config(self, settings=None):
        """
        Load the configuration either from the config file, or from the given settings.

        args:
            settings (dict): If given, the settings are pulled from this dictionary. Otherwise, the
                config file is used.
        """
        self._load_defaults()
        configfile = get_configfile()
        if settings:
            self.update(settings)
        else:
            self.update(get_appsettings(configfile))
        self.loaded = True
        self._validate()

    def _load_defaults(self):
        """Iterate over self._defaults and set all default values on self."""
        for k, v in self._defaults.items():
            self[k] = v['value']

    def _validate(self):
        """Run the validators found in self._defaults on all the corresponding values."""
        errors = []
        for k in self._defaults.keys():
            try:
                self[k] = self._defaults[k]['validator'](self[k])
            except ValueError as e:
                errors.append('\t{}: {}'.format(k, six.text_type(e)))

        if errors:
            raise ValueError(
                'Invalid config values were set: \n{}'.format('\n'.join(errors)))


config = BodhiConfig()



# -*- coding: utf-8 -*-
# Copyright 2007-2018 Red Hat, Inc. and others.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Define tools for interacting with the build system and a fake build system for development."""

from threading import Lock
import logging
import time
from functools import wraps

import koji


log = logging.getLogger('bodhi')
_buildsystem = None
# URL of the koji hub
_koji_hub = None


def multicall_enabled(func):
    """
    Decorate the given callable to enable multicall handling.

    This is used by DevBuildsys methods.

    Args:
        func (callable): The function to wrap.
    Returns:
        callable: A wrapped version of func.
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """
        If multicall is enabled, store the results from func on self.

        If multicall is not enabled, just call func and return its results as per usual.
        """
        if not self.multicall:
            return func(self, *args, **kwargs)

        # disable multicall during execution, so that inner func calls to other
        # methods don't append their results as well
        self._multicall = False
        result = func(self, *args, **kwargs)
        self.multicall_result.append([result])
        self._multicall = True
    return wrapper


class Buildsystem(object):
    """
    The parent for our buildsystem.

    Not only does this help us keep track of the functionality that we expect from our buildsystem,
    but it also alows us to create a development subclass of this object to use during development
    so we don't alter any production data.
    """

    def getBuild(self, *args, **kw):
        """Raise NotImplementedError."""
        raise NotImplementedError

    def getLatestBuilds(self, *args, **kw):
        """Raise NotImplementedError."""
        raise NotImplementedError

    def moveBuild(self, *args, **kw):
        """Raise NotImplementedError."""
        raise NotImplementedError

    def ssl_login(self, *args, **kw):
        """Raise NotImplementedError."""
        raise NotImplementedError

    def listBuildRPMs(self, *args, **kw):
        """Raise NotImplementedError."""
        raise NotImplementedError

    def listTags(self, *args, **kw):
        """Raise NotImplementedError."""
        raise NotImplementedError

    def listTagged(self, *args, **kw):
        """Raise NotImplementedError."""
        raise NotImplementedError

    def taskFinished(self, *args, **kw):
        """Raise NotImplementedError."""
        raise NotImplementedError

    def tagBuild(self, *args, **kw):
        """Raise NotImplementedError."""
        raise NotImplementedError

    def untagBuild(self, *args, **kw):
        """Raise NotImplementedError."""
        raise NotImplementedError

    def multiCall(self, *args, **kw):
        """Raise NotImplementedError."""
        raise NotImplementedError

    def getTag(self, *args, **kw):
        """Raise NotImplementedError."""
        raise NotImplementedError

    def addTag(self, *args, **kw):
        """Raise NotImplementedError."""
        raise NotImplementedError

    def deleteTag(self, *args, **kw):
        """Raise NotImplementedError."""
        raise NotImplementedError


class DevBuildsys(Buildsystem):
    """A dummy buildsystem instance used during development and testing."""

    __untag__ = []
    __moved__ = []
    __added__ = []
    __tagged__ = {}
    __rpms__ = []
    __tags__ = []

    def __init__(self):
        """Initialize the DevBuildsys."""
        self._multicall = False
        self.multicall_result = []

    @property
    def multicall(self):
        """
        Return the value of self._multicall.

        Returns:
            object: The value of self._multicall.
        """
        return self._multicall

    @multicall.setter
    def multicall(self, value):
        """
        Set the _multicall attribute to the given value.

        Args:
            value (object): The value to set the _multicall attribute to.
        """
        self._multicall = value
        self.multicall_result = []

    @classmethod
    def clear(cls):
        """Clear the state of the class variables."""
        cls.__untag__ = []
        cls.__moved__ = []
        cls.__added__ = []
        cls.__tagged__ = {}
        cls.__rpms__ = []

    def multiCall(self):
        """Emulate Koji's multiCall."""
        result = self.multicall_result
        self.multicall = False
        return result

    def moveBuild(self, from_tag, to_tag, build, *args, **kw):
        """Emulate Koji's moveBuild."""
        if to_tag is None:
            raise RuntimeError('Attempt to tag {} with None.'.format(build))
        log.debug("moveBuild(%s, %s, %s)" % (from_tag, to_tag, build))
        DevBuildsys.__moved__.append((from_tag, to_tag, build))

    def tagBuild(self, tag, build, *args, **kw):
        """Emulate Koji's tagBuild."""
        if tag is None:
            raise RuntimeError('Attempt to tag {} with None.'.format(build))
        log.debug("tagBuild(%s, %s)" % (tag, build))
        DevBuildsys.__added__.append((tag, build))

    def untagBuild(self, tag, build, *args, **kw):
        """Emulate Koji's untagBuild."""
        if tag is None:
            raise RuntimeError('Attempt to untag {} with None.'.format(build))
        log.debug("untagBuild(%s, %s)" % (tag, build))
        DevBuildsys.__untag__.append((tag, build))

    def ssl_login(self, *args, **kw):
        """Emulate Koji's ssl_login."""
        log.debug("ssl_login(%s, %s)" % (args, kw))

    def taskFinished(self, task):
        """Emulate Koji's taskFinished."""
        return True

    def getTaskInfo(self, task):
        """Emulate Koji's getTaskInfo."""
        return {'state': koji.TASK_STATES['CLOSED']}

    def getTaskRequest(self, task_id):
        """Emulate Koji's getTaskRequest."""
        return [
            u'git://pkgs.fedoraproject.org/rpms/bodhi?#2e994ca8b3296e62e8b0aadee1c5c0649559625a',
            'f17-candidate', {}]

    def listPackages(self):
        """Emulate Koji's listPackages."""
        return [
            {'package_id': 2625, 'package_name': 'nethack'},
        ]

    @multicall_enabled
    def getBuild(self, build='TurboGears-1.0.2.2-2.fc17', other=False, testing=False):
        """Emulate Koji's getBuild."""
        theid = 16058
        if other and not testing:
            theid = 16059
        elif other and testing:
            theid = 16060
        data = {'build_id': 16058,
                'completion_time': '2007-08-24 23:26:10.890319',
                'completion_ts': 1187997970,
                'creation_event_id': 151517,
                'creation_time': '2007-08-24 19:38:29.422344',
                'extra': None,
                'epoch': None,
                'id': theid,
                'owner_id': 388,
                'owner_name': 'lmacken',
                'package_id': 8,
                'state': 1,
                'tag_id': 19,
                'task_id': 127621}

        name, version, release = build.rsplit("-", 2)
        release_tokens = release.split(".")

        for token in release_tokens:
            # Starting to hardcode some dev buildsys bits for docker.
            # See https://github.com/fedora-infra/bodhi/pull/1543
            if token.endswith("container") or token.endswith("flatpak"):
                fedora_release = "f" + (token
                                        .replace("fc", "")
                                        .replace("flatpak", "")
                                        .replace("container", ""))
                tag = "%s-updates-testing" % fedora_release

                format_data = {
                    'registry': 'candidate-registry.fedoraproject.org',
                    'hash': 'sha256:2bd64a888...',
                    'version': version,
                    'release': release
                }

                if token.endswith("flatpak"):
                    format_data['repository'] = name
                else:
                    tag = "f%s-updates-testing" % token.replace("fc", "").replace("container", "")
                    format_data['repository'] = "{}/{}".format(fedora_release, name)

                data['extra'] = {
                    'container_koji_task_id': 19708268,
                    'image': {
                        'index': {
                            'pull': ['{registry}/{repository}@sha256:{hash}'
                                     .format(**format_data),
                                     '{registry}/{repository}:{version}-{release}'
                                     .format(**format_data)],
                        }
                    },
                }

                if token.endswith("flatpak"):
                    data['extra']['image']['flatpak'] = True

                break

            # Hardcoding for modules in the dev buildsys
            if token.startswith("2017"):
                tag = "f27M-updates-testing"
                data['extra'] = {
                    'typeinfo': {'module': {'more': 'mbs stuff goes here'}}
                }
                break

            if token.startswith("fc"):
                if testing:
                    tag = "f%s-updates-testing" % token.replace("fc", "")
                    break
                else:
                    tag = "f%s-updates-candidate" % token.replace("fc", "")
                    break

            if token.startswith("el"):
                tag = "dist-%sE-epel-testing-candidate" % token.replace("el", "")
                break
        else:
            raise ValueError("Couldn't determine dist for build '%s'" % build)

        if other:
            if testing:
                release_tokens[0] = str(int(release_tokens[0]) + 2)
            else:
                release_tokens[0] = str(int(release_tokens[0]) + 1)
            release = ".".join(release_tokens)
            build = "%s-%s-%s" % (name, version, release)

        data.update({'name': name,
                     'nvr': build,
                     'package_name': name,
                     'release': release,
                     'tag_name': tag,
                     'version': version})

        return data

    def listBuildRPMs(self, id, *args, **kw):
        """Emulate Koji's listBuildRPMs."""
        rpms = [{'arch': 'src',
                 'build_id': 6475,
                 'buildroot_id': 1883,
                 'buildtime': 1178868422,
                 'epoch': None,
                 'id': 62330,
                 'name': 'TurboGears',
                 'nvr': 'TurboGears-1.0.2.2-2.fc17',
                 'payloadhash': '6787febe92434a9be2a8f309d0e2014e',
                 'release': '2.fc17',
                 'size': 761742,
                 'version': '1.0.2.2'},
                {'arch': 'noarch',
                 'build_id': 6475,
                 'buildroot_id': 1883,
                 'buildtime': 1178868537,
                 'epoch': None,
                 'id': 62331,
                 'name': 'TurboGears',
                 'nvr': 'TurboGears-1.0.2.2-2.fc17',
                 'payloadhash': 'f3ec9bdce453816f94283a15a47cb952',
                 'release': '2.fc17',
                 'size': 1993385,
                 'version': '1.0.2.2'},
                ]
        if id == 16059:  # for updateinfo.xml tests
            rpms[0]['nvr'] = rpms[1]['nvr'] = 'TurboGears-1.0.2.2-3.fc17'
            rpms[0]['release'] = rpms[1]['release'] = '3.fc17'
        rpms += DevBuildsys.__rpms__
        return rpms

    def listTags(self, build, *args, **kw):
        """Emulate Koji's listTags."""
        if 'el5' in build or 'el6' in build:
            release = build.split('.')[-1].replace('el', '')
            result = [
                {'arches': 'i386 x86_64 ppc ppc64', 'id': 10, 'locked': True,
                 'name': 'dist-%sE-epel-testing-candidate' % release, 'perm': None,
                 'perm_id': None},
                {'arches': 'i386 x86_64 ppc ppc64', 'id': 10, 'locked': True,
                 'name': 'dist-%sE-epel-testing-candidate' % release, 'perm': None,
                 'perm_id': None},
                {'arches': 'i386 x86_64 ppc ppc64', 'id': 5, 'locked': True,
                 'name': 'dist-%sE-epel' % release, 'perm': None, 'perm_id': None}]
        elif 'el7' in build:
            release = build.split('.')[-1].replace('el', 'epel')
            result = [
                {'arches': 'i386 x86_64 ppc ppc64', 'id': 10, 'locked': True,
                 'name': '%s-testing-candidate' % release, 'perm': None, 'perm_id': None},
                {'arches': 'i386 x86_64 ppc ppc64', 'id': 5, 'locked': True, 'name': '%s' % release,
                 'perm': None, 'perm_id': None},
                {'arches': 'i386 x86_64 ppc ppc64', 'id': 5, 'locked': True,
                 'name': '%s-testing' % release, 'perm': None, 'perm_id': None}]
        elif '-master-' in build or build.startswith(('nodejs-6-', 'nodejs-8-', 'nodejs-9-')):
            # Hardcoding for modules in the dev buildsys
            result = [
                {'arches': 'x86_64', 'id': 15, 'locked': True,
                 'name': 'f27M-updates-candidate'},
                {'arches': 'x86_64', 'id': 16, 'locked': True,
                 'name': 'f27M-updates-testing'},
                {'arches': 'x86_64', 'id': 17, 'locked': True,
                 'name': 'f27M'},
            ]
        elif 'container' in build:
            result = [
                {'arches': 'x86_64', 'id': 15, 'locked': True,
                 'name': 'f28C-updates-candidate'},
                {'arches': 'x86_64', 'id': 16, 'locked': True,
                 'name': 'f28C-updates-testing'},
                {'arches': 'x86_64', 'id': 17, 'locked': True,
                 'name': 'f28C'},
            ]
        elif 'flatpak' in build:
            result = [
                {'arches': 'x86_64', 'id': 15, 'locked': True,
                 'name': 'f28F-updates-candidate'},
                {'arches': 'x86_64', 'id': 16, 'locked': True,
                 'name': 'f28F-updates-testing'},
                {'arches': 'x86_64', 'id': 17, 'locked': True,
                 'name': 'f28F'},
            ]
        else:
            release = build.split('.')[-1].replace('fc', 'f')
            result = [
                {'arches': 'i386 x86_64 ppc ppc64', 'id': 10, 'locked': True,
                 'name': '%s-updates-candidate' % release, 'perm': None, 'perm_id': None},
                {'arches': 'i386 x86_64 ppc ppc64', 'id': 5, 'locked': True, 'name': '%s' % release,
                 'perm': None, 'perm_id': None},
                {'arches': 'i386 x86_64 ppc ppc64', 'id': 5, 'locked': True,
                 'name': '%s-updates-testing' % release, 'perm': None, 'perm_id': None}]
        if build in DevBuildsys.__tagged__:
            for tag in DevBuildsys.__tagged__[build]:
                result += [{'name': tag}]
        return result

    @multicall_enabled
    def listTagged(self, tag, *args, **kw):
        """List updates tagged with teh given tag."""
        builds = []
        for build in [self.getBuild(),
                      self.getBuild(other=True),
                      self.getBuild(other=True, testing=True)]:
            if build['nvr'] in self.__untag__:
                log.debug('Pruning koji build %s' % build['nvr'])
                continue
            elif build['tag_name'] == tag:
                builds.append(build)
        for build in DevBuildsys.__tagged__:
            for tag_ in DevBuildsys.__tagged__[build]:
                if tag_ == tag:
                    builds.append(self.getBuild(build))
        log.debug(builds)
        return builds

    def getLatestBuilds(self, *args, **kw):
        """
        Return a list of the output from self.getBuild().

        Returns:
            list: A list of the latest builds from getBuild().
        """
        return [self.getBuild()]

    def getTag(self, taginfo, **kw):
        """
        Retrieve the given tag from koji.

        Args:
            taginfo (int or basestring): The tag you want info about.
            strict (bool): If True, raise an Exception if epel tags are queried. Defaults to False.
        Returns:
            dict or None: A dictionary of tag information, or None if epel is requested and strict
                is False.
        Raises:
            koji.GenericError: If strict is True and epel is requested.
        """
        for nr in self.__tags__:
            if taginfo == self.__tags__[nr][0]:
                toreturn = self.__tags__[nr][1]
                toreturn['id'] = nr
                return toreturn

        if isinstance(taginfo, int):
            taginfo = "f%d" % taginfo

        if taginfo.startswith("epel"):
            if kw.get("strict", False):
                raise koji.GenericError("Invalid tagInfo: '%s'" % taginfo)

            else:
                return None

        return {'maven_support': False, 'locked': False, 'name': taginfo,
                'perm': None, 'id': 246, 'arches': None,
                'maven_include_all': False, 'perm_id': None}

    def addTag(self, tag, **opts):
        """Emulate tag adding."""
        if 'parent' not in opts:
            raise ValueError('No parent in tag options')
        for nr in self.__tags__:
            if self.__tags__[0] == tag:
                raise ValueError('Tag %s already exists' % tag)
        opts['locked'] = False
        opts['maven_support'] = False
        opts['name'] = tag
        opts['perm'] = 'admin'
        opts['arches'] = None
        opts['maven_include_all'] = False
        opts['perm_id'] = 1
        self.__tags__.append((tag, opts))

    def deleteTag(self, tagid):
        """Emulate tag deletion."""
        del self.__tags__[tagid]

    def getRPMHeaders(self, rpmID, headers):
        """
        Return headers for the given RPM.

        Args:
            rpmID (basestring): The RPM you want headers for.
            headers (object): Unused.
        Returns:
            dict: A dictionary of RPM headers.
        """
        if rpmID == 'raise-exception.src':
            raise Exception
        elif rpmID == 'do-not-find-anything.src':
            return None
        else:
            headers = {
                'description': (
                    "The libseccomp library provides an easy to use interface to the "
                    "Linux Kernel's\nsyscall filtering mechanism, seccomp. The "
                    "libseccomp API allows an application\nto specify which "
                    "syscalls, and optionally which syscall arguments, the\n"
                    "application is allowed to execute, all of which are "
                    "enforced by the Linux\nKernel."),
                'url': 'http://libseccomp.sourceforge.net',
                'changelogname': [
                    'Fedora Release Engineering <rel-eng@lists.fedoraproject.org> - 2.1.0-1',
                    'Paul Moore <pmoore@redhat.com> - 2.1.0-0',
                    'Paul Moore <pmoore@redhat.com> - 2.0.0-0',
                    'Paul Moore <pmoore@redhat.com> - 1.0.1-0',
                    'Paul Moore <pmoore@redhat.com> - 1.0.0-0',
                    'Fedora Release Engineering <rel-eng@lists.fedoraproject.org> - 0.1.0-2',
                    'Paul Moore <pmoore@redhat.com> - 0.1.0-1',
                    'Paul Moore <pmoore@redhat.com> - 0.1.0-0'],
                'summary': 'Enhanced seccomp library',
                'version': '2.1.0',
                'changelogtime': [
                    1375531200, 1370952000, 1359374400, 1352808000, 1343736000,
                    1342699200, 1341921600, 1339502400],
                'changelogtext': [
                    '- Rebuilt for https://fedoraproject.org/wiki/Fedora_20_Mass_Rebuild',
                    '- New upstream version\n- Added support for the ARM architecture\n'
                    '- Added the scmp_sys_resolver tool',
                    '- New upstream version',
                    '- New upstream version with several important fixes',
                    '- New upstream version\n- Remove verbose build patch as it is no '
                    'longer needed\n- Enable _smp_mflags during build stage',
                    '- Rebuilt for https://fedoraproject.org/wiki/Fedora_18_Mass_Rebuild',
                    '- Limit package to x86/x86_64 platforms (RHBZ #837888)',
                    '- Initial version'],
                'release': '1.fc20',
                'name': 'libseccomp'
            }
            if rpmID == "TurboGears-2.0.0.0-1.fc17.src":
                headers['changelogname'].insert(0, 'Randy Barlow <bowlofeggs@fp.o> - 2.2.0-1')
                headers['changelogtext'].insert(0, '- Added some bowlofeggs charm.')
                headers['changelogtime'].insert(0, 1375531201)
            elif rpmID == 'TurboGears-1.9.1-1.fc17.src':
                headers['changelogtext'] = []
            elif rpmID == 'TurboGears-1.9.1-42.fc17.src':
                # Make sure only a single changelog entry is present
                headers['changelogname'] = ['Randy Barlow <bowlofeggs@fp.o> - 1.9.1-42']
                headers['changelogtext'] = ["- Hope I didn't break anything!"]
                headers['changelogtime'] = [1375531200]
            return headers


def koji_login(config, authenticate):
    """
    Login to Koji and return the session.

    Args:
        config (bodhi.server.config.BodhiConfig): Bodhi's configuration dictionary.
        authenticate (bool): If True, establish an authenticated client session.
    Returns:
        koji.ClientSession: An authenticated Koji ClientSession that is ready to use.
    """
    koji_options = {
        'krb_rdns': False,
        'max_retries': 30,
        'retry_interval': 10,
        'offline_retry': True,
        'offline_retry_interval': 10,
        'anon_retry': True,
    }

    koji_client = koji.ClientSession(_koji_hub, koji_options)
    if authenticate and not koji_client.krb_login(**get_krb_conf(config)):
        log.error('Koji krb_login failed')
    return koji_client


def get_krb_conf(config):
    """
    Return arguments for krb_login.

    Args:
        config (bodhi.server.config.BodhiConfig): Bodhi's configuration dictionary.
    Returns:
        dict: A dictionary containing three keys:
            principal: The kerberos principal to use.
            keytab: The kerberos keytab to use.
            ccache: The kerberos ccache to use.
    """
    principal = config.get('krb_principal')
    keytab = config.get('krb_keytab')
    ccache = config.get('krb_ccache')
    args = {}
    if principal:
        args['principal'] = principal
    if keytab:
        args['keytab'] = keytab
    if ccache:
        args['ccache'] = ccache
    return args


def get_session():
    """
    Get a new buildsystem instance.

    Returns:
        koji.ClientSession or DevBuildsys: A buildsystem client instance.
    Raises:
        RuntimeError: If the build system has not been initialized. See setup_buildsystem().
    """
    global _buildsystem, _buildsystem_login_lock
    if _buildsystem is None:
        raise RuntimeError('Buildsys needs to be setup')
    with _buildsystem_login_lock:
        return _buildsystem()


def teardown_buildsystem():
    """Tear down the build system."""
    global _buildsystem
    _buildsystem = None
    DevBuildsys.clear()


def setup_buildsystem(settings, authenticate=True):
    """
    Initialize the buildsystem client.

    Args:
        settings (bodhi.server.config.BodhiConfig): Bodhi's config.
        authenticate (bool): If True, establish an authenticated Koji session. Defaults to True.
    Raises:
        ValueError: If the buildsystem is configured to an invalid value.
    """
    global _buildsystem, _koji_hub, _buildsystem_login_lock
    if _buildsystem:
        return

    _buildsystem_login_lock = Lock()
    _koji_hub = settings.get('koji_hub')
    buildsys = settings.get('buildsystem')

    if buildsys == 'koji':
        log.debug('Using Koji Buildsystem')

        def get_koji_login():
            """Call koji_login with settings and return the result."""
            return koji_login(config=settings, authenticate=authenticate)

        _buildsystem = get_koji_login
    elif buildsys in ('dev', 'dummy', None):
        log.debug('Using DevBuildsys')
        _buildsystem = DevBuildsys
    else:
        raise ValueError('Buildsys %s not known' % buildsys)


def wait_for_tasks(tasks, session=None, sleep=300):
    """
    Wait for a list of koji tasks to complete.

    Args:
        tasks (list): The return value of Koji's multiCall().
        session (koji.ClientSession or None): A Koji client session to use. If not provided, the
            function will acquire its own session.
        sleep (int): How long to sleep between polls on Koji when waiting for tasks to complete.
    Returns:
        list: A list of failed tasks. An empty list indicates that all tasks completed successfully.
    """
    log.debug("Waiting for %d tasks to complete: %s" % (len(tasks), tasks))
    failed_tasks = []
    if not session:
        session = get_session()
    for task in tasks:
        if not task:
            log.debug("Skipping task: %s" % task)
            continue
        while not session.taskFinished(task):
            time.sleep(sleep)
        task_info = session.getTaskInfo(task)
        if task_info['state'] != koji.TASK_STATES['CLOSED']:
            log.error("Koji task %d failed" % task)
            failed_tasks.append(task)
    log.debug("%d tasks completed successfully, %d tasks failed." % (
        len(tasks) - len(failed_tasks), len(failed_tasks)))
    return failed_tasks



# -*- coding: utf-8 -*-
# Copyright © 2018 Red Hat, Inc.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Copy containers between registries.

This is a very limited version of the skopeo tool, but with support
for manifests lists and OCI image indexes.
https://github.com/containers/image/pull/400 will make this
unnecessary.

The only subcommand that is supported is 'copy', and the only supported image references
are Docker registry references of the form 'docker://docker-reference'.

No global options are supported, and only selected options to 'copy' are supported (see
--help for details.)

Some other things that aren't implemented (but could be added if necessary):
 - Handling of www-authenticate responses, necessary to log in to docker.io
 - Special handling of 'docker.io' as 'registry-1.docker.io'
 - Reading ~/.docker/config.json or $XDG_RUNTIME_DIR/containers/auth.json
 - Handling foreign layers
"""

import json
import logging
import os
import shutil
import tempfile

import click
import requests
from requests.exceptions import SSLError, ConnectionError
from six.moves.urllib.parse import urlparse, urlunparse


@click.group()
def main():
    """Simplified Skopeo work-alike with manifest list support."""
    pass  # pragma: no cover


logger = logging.getLogger('skopeo-lite')
logging.basicConfig(level=logging.INFO)


MEDIA_TYPE_MANIFEST_V2 = 'application/vnd.docker.distribution.manifest.v2+json'
MEDIA_TYPE_LIST_V2 = 'application/vnd.docker.distribution.manifest.list.v2+json'
MEDIA_TYPE_OCI = 'application/vnd.oci.image.manifest.v1+json'
MEDIA_TYPE_OCI_INDEX = 'application/vnd.oci.image.index.v1+json'


class RegistrySpec(object):
    """Information about a docker registry/repository/tag as specified on the command line."""

    def __init__(self, registry, repo, tag, creds, tls_verify, cert_dir):
        """
        Initialize the registry spec.

        Args:
           registry (str): The hostname.
           repo (str): The repository name within the registry.
           tag (str): The tag within the repository.
           creds (str): user:password (may be None).
           tls_verify (bool): If True, HTTPS with a verified certificate is required.
           cert_dir (str): A path to directory holding client certificates, or None.
        """
        self.registry = registry
        self.repo = repo
        self.tag = tag
        self.creds = creds
        self.cert_dir = cert_dir
        self.tls_verify = tls_verify

    def get_session(self):
        """Create a RegistrySession object for the spec."""
        return RegistrySession(self.registry, insecure=not self.tls_verify,
                               creds=self.creds, cert_dir=self.cert_dir)

    def get_endpoint(self):
        """Create a RegistryEndpoint object for the spec."""
        return RegistryEndpoint(self)


def parse_spec(spec, creds, tls_verify, cert_dir):
    """
    Parse a string into a RegistrySpec, adding extra information.

    Args:
       spec (str): docker://<hostname>/<repo>[:tag] - latest will be used
           if tag is omitted.
       creds (str): user:password (may be None).
       tls_verify (bool): If True https with a verified certificate is required
       cert_dir (str): A path to directory holding client certificates, or None.
    Returns:
       RegistrySpec: The resulting registry spec.
    Raises:
       click.BadArgumentUsage: If the string cannot be parsed
    """
    if spec.startswith('docker:'):
        if not spec.startswith('docker://'):
            raise click.BadArgumentUsage(
                "Registry specification should be docker://REGISTRY/PATH[:TAG]")

        parts = spec[len('docker://'):].split('/', 1)
        if len(parts) == 1:
            raise click.BadArgumentUsage(
                "Registry specification should be docker://REGISTRY/PATH[:TAG]")

        registry, path = parts
        parts = path.split(':', 1)
        if len(parts) == 1:
            repo, tag = parts[0], 'latest'
        else:
            repo, tag = parts

        return RegistrySpec(registry, repo, tag, creds, tls_verify, cert_dir)
    else:
        raise click.BadArgumentUsage("Unknown source/destination: {}".format(spec))


class RegistrySession(object):
    """Wrapper around requests.Session adding docker-specific behavior."""

    def __init__(self, registry, insecure=False, creds=None, cert_dir=None):
        """
        Initialize the RegistrySession.

        Args:
            registry (str): The hostname of the registry.
            insecure (bool): If True don't verify TLS certificates and fallback to HTTP.
            creds (str): user:password (may be None).
            cert_dir (str): A path to directory holding client certificates, or None.
        """
        self.registry = registry
        self._resolved = None
        self.insecure = insecure

        self.cert = self._find_cert(cert_dir)

        self.auth = None
        if creds is not None:
            username, password = creds.split(':', 1)
            self.auth = requests.auth.HTTPBasicAuth(username, password)

        self._fallback = None
        self._base = 'https://{}'.format(self.registry)
        if insecure:
            # In the insecure case, if the registry is just a hostname:port, we
            # don't know whether to talk HTTPS or HTTP to it, so we try first
            # with https then fallback
            self._fallback = 'http://{}'.format(self.registry)

        self.session = requests.Session()

    def _find_cert_dir(self):
        """
        Return a path to a directory containing TLS client certificates to use for authentication.

        Returns:
            str or None: If a path is found, it is returned. Otherwise None is returned.
        """
        hostport = self.registry

        for d in ('/etc/containers/certs.d', '/etc/docker/certs.d'):
            certs_dir = os.path.join(d, hostport)
            if os.path.isdir(certs_dir):
                return certs_dir

        return None

    def _find_cert(self, cert_dir):
        """
        Return a TLS client certificate to be used to authenticate to servers.

        Args:
            cert_dir (str or None): A directory to look for certs in. None indicates to use
               find_cert_dir() to find the path. Defaults to None.
        Returns:
            tuple or None: If no certificate is found, None is returned, otherwise, a 2-tuple
               is returned, the first element is the path to a certificate, the second element
               is the path to the matching key.
        Raises:
            RuntimeError: If a key is found without a matching certificate or vice versa.
        """
        if cert_dir is None:
            cert_dir = self._find_cert_dir()

        if cert_dir is None:
            return None

        for l in sorted(os.listdir(cert_dir)):
            if l.endswith('.cert'):
                certpath = os.path.join(cert_dir, l)
                keypath = certpath[:-5] + '.key'
                if not os.path.exists(keypath):
                    raise RuntimeError("Cannot find key file for {}".format(certpath))
                return (certpath, keypath)
            elif l.endswith('.key'):
                # Should have found <x>.cert first
                keypath = os.path.join(cert_dir, l)
                raise RuntimeError("Cannot find certificate file for {}".format(keypath))

        return None

    def _wrap_method(self, f, relative_url, *args, **kwargs):
        """
        Perform an HTTP request with appropriate options and fallback handling.

        This is used to implement methods like get, head, etc. It modifies
        kwargs, tries to do the operation, then if a TLS request fails and
        TLS validation is not required, tries again with a non-TLS URL.

        Args:
            f (callable): callback to actually perform the request.
            relative_url (str): URL relative to the toplevel hostname.
            kwargs: Additional arguments passed to requests.Session.get.
        Returns:
            requests.Response: The response object.
        """
        kwargs['auth'] = self.auth
        kwargs['cert'] = self.cert
        kwargs['verify'] = not self.insecure
        res = None
        if self._fallback:
            try:
                res = f(self._base + relative_url, *args, **kwargs)
                self._fallback = None  # don't fallback after one success
            except (SSLError, ConnectionError):
                self._base = self._fallback
                self._fallback = None
        if res is None:
            res = f(self._base + relative_url, *args, **kwargs)
        return res

    def get(self, relative_url, **kwargs):
        """
        Do a HTTP GET.

        Args:
            relative_url (str): URL relative to the toplevel hostname.
            kwargs: Additional arguments passed to requests.Session.get.
        Returns:
            requests.Response: The response object.
        """
        return self._wrap_method(self.session.get, relative_url, **kwargs)

    def head(self, relative_url, data=None, **kwargs):
        """
        Do a HTTP HEAD.

        Args:
            relative_url (str): URL relative to the toplevel hostname.
            kwargs: Additional arguments passed to requests.Session.head.
        Returns:
            requests.Response: The response object.
        """
        return self._wrap_method(self.session.head, relative_url, **kwargs)

    def post(self, relative_url, data=None, **kwargs):
        """
        Do a HTTP POST.

        Args:
            relative_url (str): URL relative to the toplevel hostname.
            data: Data to include with the post, as for requests.SESSION.
            kwargs: Additional arguments passed to requests.Session.post.
        Returns:
            requests.Response: The response object.
        """
        return self._wrap_method(self.session.post, relative_url, data=data, **kwargs)

    def put(self, relative_url, data=None, **kwargs):
        """
        Do a HTTP PUT.

        Args:
            relative_url (str): URL relative to the toplevel hostname.
            data: Data to include with the put, as for requests.SESSION.
            kwargs: Additional arguments passed to requests.Session.put.
        Returns:
            requests.Response: The response object.
        """
        return self._wrap_method(self.session.put, relative_url, data=data, **kwargs)


class ManifestInfo(object):
    """Information about a manifest downloaded from the registry."""

    def __init__(self, contents, digest, media_type, size):
        """
        Initialize the ManifestInfo.

        Args:
            contents (bytes): The contents.
            media_type (str): The type of the content.
            digest (str): The digest of the content.
            size (int): The size of the download, in bytes.
        """
        self.contents = contents
        self.digest = digest
        self.media_type = media_type
        self.size = size


def get_manifest(session, repository, ref):
    """
    Download a manifest from a registry.

    Args:
        session (RegistrySession): The session object.
        repository (str): The repository to download from.
        ref (str): A digest, or a tag.
    Returns:
        ManifestInfo: Information about the downloaded content.
    """
    logger.debug("%s: Retrieving manifest for %s:%s", session.registry, repository, ref)

    headers = {
        'Accept': ', '.join((
            MEDIA_TYPE_MANIFEST_V2,
            MEDIA_TYPE_LIST_V2,
            MEDIA_TYPE_OCI,
            MEDIA_TYPE_OCI_INDEX
        ))
    }

    url = '/v2/{}/manifests/{}'.format(repository, ref)
    response = session.get(url, headers=headers)
    response.raise_for_status()
    return ManifestInfo(response.content,
                        response.headers['Docker-Content-Digest'],
                        response.headers['Content-Type'],
                        int(response.headers['Content-Length']))


class DirectoryEndpoint(object):
    """
    The source or destination of a copy operation to a local directory.

    This is used only as a local intermediate, and for simplicity, the storage format is
    not exactly the same as for a dir:// reference as understood by skopeo.
    """

    def __init__(self, directory):
        """
        Initialize the DirectoryEndpoint.

        Args:
            directory (str): The path to the directory.
        """
        self.directory = directory

    def start_write(self):
        """Do setup before writing to the endpoint."""
        with open(os.path.join(self.directory, 'oci-layout'), 'w') as f:
            f.write('{"imageLayoutVersion": "1.0.0"}\n')

    def get_blob_path(self, digest):
        """
        Get the path that a blob with the given digest would be stored at.

        Args:
            digest (str): The digest of a blob to be stored.
        Returns:
            str: The full path where the blob would be stored.
        """
        algorithm, digest = digest.split(':', 2)
        return os.path.join(self.directory, 'blobs', algorithm, digest)

    def ensure_blob_path(self, digest):
        """
        Get path for a blob, creating parent directories if necessary.

        Args:
            digest (str): The digest of a blob to be stored.
        Returns:
            str: The full path where the blob would be stored.
        """
        path = self.get_blob_path(digest)

        parent = os.path.dirname(path)
        if not os.path.exists(parent):
            os.makedirs(parent)

        return path

    def get_blob(self, digest):
        """
        Return the contents of a blob object.

        Args:
            digest (str): The digest to retrieve.
        Returns:
            bytes: The contents of the blob.
        """
        with open(self.get_blob_path(digest), 'rb') as f:
            return f.read()

    def has_blob(self, digest):
        """
        Check if the repository has a blob with the given digest.

        Args:
            digest (str): The digest to check for.
        Returns:
            bool: True if the blob exists.
        """
        return os.path.exists(self.get_blob_path(digest))

    def write_blob(self, digest, contents):
        """
        Save a blob to the directory endpoint.

        Args:
            digest (str): The digest of the blob's contents.
            contents (bytes): The contents of the blob.
        """
        path = self.ensure_blob_path(digest)

        with open(path, 'wb') as f:
            f.write(contents)

    def get_manifest(self, digest=None, media_type=None):
        """
        Get a manifest from the endpoint.

        Arguments:
            digest (str): The digest of manifest to retrieve, or None to get the
               main manifest for the endpoint.
            media_type (str): The expected media type of the manifest.
        Returns:
            ManifestInfo: An object containing the contents of the manifest
               and other relevant information.
        """
        if digest is not None:
            contents = self.get_blob(digest)
        else:
            manifest_path = os.path.join(self.directory, 'manifest.json')
            if os.path.exists(manifest_path):
                with open(manifest_path, 'rb') as f:
                    contents = f.read()
                parsed = json.loads(contents)
                media_type = parsed.get('mediaType', MEDIA_TYPE_OCI)
            else:
                index_path = os.path.join(self.directory, 'index.json')
                with open(index_path, 'rb') as f:
                    contents = f.read()
                parsed = json.loads(contents)
                media_type = parsed.get('mediaType', MEDIA_TYPE_OCI_INDEX)

        return ManifestInfo(contents, digest, media_type, len(contents))

    def write_manifest(self, info, toplevel=False):
        """
        Store a manifest to the endpoint.

        Args:
            info (ManifestInfo): An object containing the contents of the manifest
                and other relevant information.
            toplevel (bool): If True, this should be the main manifest stored
                in the endpoint.
        """
        if not toplevel:
            self.write_blob(info.digest, info.contents)
        elif info.media_type in (MEDIA_TYPE_LIST_V2, MEDIA_TYPE_OCI_INDEX):
            with open(os.path.join(self.directory, 'index.json'), 'wb') as f:
                f.write(info.contents)
        else:
            with open(os.path.join(self.directory, 'manifest.json'), 'wb') as f:
                f.write(info.contents)


class RegistryEndpoint(object):
    """The source or destination of a copy operation to a docker registry."""

    def __init__(self, spec):
        """
        Initialize the RegistryEndpoint.

        Args:
            spec (RegistrySpec): A specification of registry, repository and tag.
        """
        self.session = spec.get_session()
        self.registry = spec.registry
        self.repo = spec.repo
        self.tag = spec.tag

    def start_write(self):
        """Do setup before writing to the endpoint."""
        pass

    def download_blob(self, digest, size, blob_path):
        """
        Download a blob from the registry to a local file.

        Args:
            digest (str): The digest of the blob to download.
            size (int): The size of blob.
            blob_path (str): The local path to write the blob to.
        """
        logger.info("%s: Downloading %s (size=%s)", self.registry, blob_path, size)

        url = "/v2/{}/blobs/{}".format(self.repo, digest)
        result = self.session.get(url, stream=True)
        result.raise_for_status()

        try:
            with open(blob_path, 'wb') as f:
                for block in result.iter_content(10 * 1024):
                    f.write(block)
        finally:
            result.close()

    def upload_blob(self, digest, size, blob_path):
        """
        Upload a blob from a local file to the registry.

        Args:
            digest (str): The digest of the blob to upload.
            size (int): The size of blob to upload.
            blob_path (str): The local path to read the blob from.
        """
        logger.info("%s: Uploading %s (size=%s)", self.registry, blob_path, size)

        url = "/v2/{}/blobs/uploads/".format(self.repo)
        result = self.session.post(url, data='')
        result.raise_for_status()

        if result.status_code != requests.codes.ACCEPTED:
            # if it was a failed response 4xx or 5xx then the raise_for_status()
            # would have raised - so it's a "successful" response - but by the docker v2 api,
            # a 202 ACCEPTED should be found here, not a 200 or other successful response.
            raise RuntimeError("Unexpected successful response %s (202 expected)",
                               result.status_code)

        upload_url = result.headers.get('Location')
        parsed = urlparse(upload_url)
        if parsed.query == '':
            query = 'digest=' + digest
        else:
            query = parsed.query + '&digest=' + digest
        relative = urlunparse(('', '', parsed.path, parsed.params, query, ''))

        headers = {
            'Content-Length': str(size),
            'Content-Type': 'application/octet-stream'
        }
        with open(blob_path, 'rb') as f:
            result = self.session.put(relative, data=f, headers=headers)

        result.raise_for_status()
        if result.status_code != requests.codes.CREATED:
            # if it was a failed response 4xx or 5xx then the raise_for_status()
            # would have raised - so it's a "successful" response - but by the docker v2 api,
            # a 202 CREATED should be found here, not a 200 or other successful response.
            raise RuntimeError("Unexpected successful response %s (201 expected)",
                               result.status_code)

    def link_blob(self, digest, src_repo):
        """
        Create a new reference to an object in another repository on the same registry.

        By using the "mount" operation from the docker protocol, we avoid having
        to download and upload data.

        Args:
            digest (str): The digest of the blob to create a new reference to.
            src_repo (str): Another repository in the same registry which already has a
               blob with the given digest.
        """
        logger.info("%s: Linking blob %s from %s to %s",
                    self.registry, digest, src_repo, self.repo)

        # Check that it exists in the source repository
        url = "/v2/{}/blobs/{}".format(src_repo, digest)
        result = self.session.head(url)
        result.raise_for_status()

        url = "/v2/{}/blobs/uploads/?mount={}&from={}".format(self.repo, digest, src_repo)
        result = self.session.post(url, data='')
        result.raise_for_status()

        if result.status_code != requests.codes.CREATED:
            # A 202-Accepted would mean that the source blob didn't exist and
            # we're starting an upload - but we've checked that above
            raise RuntimeError("Blob mount had unexpected status {}".format(result.status_code))

    def has_blob(self, digest):
        """
        Check if the repository has a blob with the given digest.

        Args:
            digest (str): The digest to check for.
        Returns:
            bool: True if the blob exists.
        """
        url = "/v2/{}/blobs/{}".format(self.repo, digest)
        result = self.session.head(url, stream=True)
        if result.status_code == 404:
            return False
        result.raise_for_status()
        return True

    def get_manifest(self, digest=None, media_type=None):
        """
        Get a manifest from the endpoint.

        Args:
            digest (str): The digest of manifest to retrieve, or None to get the
               main manifest for the endpoint.
            media_type (str): The expected media type of the manifest.
        Returns:
            ManifestInfo: An object containing the contents of the manifest
               and other relevant information.
        """
        if digest is None:
            return get_manifest(self.session, self.repo, self.tag)
        else:
            return get_manifest(self.session, self.repo, digest)

    def write_manifest(self, info, toplevel=False):
        """
        Store a manifest to the endpoint.

        Args:
            info (ManifestInfo): An object containing the contents of the manifest
                and other relevant information.
            toplevel (bool): If True, this should be the main manifest stored
                in the endpoint.
        """
        if toplevel:
            ref = self.tag
        else:
            ref = info.digest

        logger.info("%s: Storing manifest as %s", self.registry, ref)

        url = '/v2/{}/manifests/{}'.format(self.repo, ref)
        headers = {'Content-Type': info.media_type}
        response = self.session.put(url, data=info.contents, headers=headers)
        response.raise_for_status()


class Copier(object):
    """
    Implements a copy operation between different endpoints.

    As currently implemented, only copying to/from a registry and a directory
    or within the *same* registry is implemented.
    """

    def __init__(self, src, dest):
        """Initialize the Copier.

        Args:
            src (str): The source endpoint.
            dest (str): The destination endpoint.
        """
        self.src = src
        self.dest = dest

    def _copy_blob(self, digest, size):
        """
        Copy a blob with given digest and size from the source to the destination.

        Args:
            digest (str): The digest of the blob to copy.
            size (int): The size of the blob to copy.
        """
        if self.dest.has_blob(digest):
            return

        if isinstance(self.src, RegistryEndpoint) and isinstance(self.dest, DirectoryEndpoint):
            self.src.download_blob(digest, size, self.dest.ensure_blob_path(digest))
        elif isinstance(self.src, DirectoryEndpoint) and isinstance(self.dest, RegistryEndpoint):
            self.dest.upload_blob(digest, size, self.src.get_blob_path(digest))
        elif isinstance(self.src, RegistryEndpoint) and isinstance(self.dest, RegistryEndpoint):
            if self.src.registry == self.dest.registry:
                self.dest.link_blob(digest, self.src.repo)
        # Other forms of copying are not needed currently, and not implemented

    def _copy_manifest(self, info, toplevel=False):
        """
        Copy the manifest referenced by info from the source to destination.

        Args:
            info (ManifestInfo): References the manifest to be copied.
            toplevel (bool): If True, this should be the main manifest referenced in the repository.
                Defaults to False.
        Raises:
            RuntimeError: If the referenced media type is not supported by this client.
        """
        references = []
        if info.media_type in (MEDIA_TYPE_MANIFEST_V2, MEDIA_TYPE_OCI):
            manifest = json.loads(info.contents)
            references.append((manifest['config']['digest'], manifest['config']['size']))
            for layer in manifest['layers']:
                references.append((layer['digest'], layer['size']))
        else:
            raise RuntimeError("Unhandled media type %s", info.media_type)

        for digest, size in references:
            self._copy_blob(digest, size)

        self.dest.write_manifest(info, toplevel=toplevel)

    def copy(self):
        """Perform the copy operation."""
        self.dest.start_write()
        info = self.src.get_manifest()
        if info.media_type in (MEDIA_TYPE_MANIFEST_V2, MEDIA_TYPE_OCI):
            self._copy_manifest(info, toplevel=True)
        elif info.media_type in (MEDIA_TYPE_LIST_V2, MEDIA_TYPE_OCI_INDEX):
            manifest = json.loads(info.contents)
            for m in manifest['manifests']:
                referenced = self.src.get_manifest(digest=m['digest'], media_type=m['mediaType'])
                self._copy_manifest(referenced)
            self.dest.write_manifest(info, toplevel=True)

        else:
            raise RuntimeError("Unhandled media type %s", info.media_type)


@main.command()
@click.option('--src-creds', '--screds', 'src_creds', metavar='USERNAME[:PASSWORD]',
              help='Use USERNAME[:PASSWORD] for accessing the source registry')
@click.option('--src-tls-verify', type=bool, default=True,
              help=('require HTTPS and verify certificates when talking to the '
                    'container source registry (defaults to true)'))
@click.option('--src-cert-dir', metavar='PATH',
              help=('use certificates at PATH (*.crt, *.cert, *.key) to connect to the '
                    'source registry'))
@click.option('--dest-creds', '--dcreds', 'dest_creds', metavar='USERNAME[:PASSWORD]',
              help='Use USERNAME[:PASSWORD] for accessing the destination registry')
@click.option('--dest-tls-verify', type=bool, default=True,
              help=('require HTTPS and verify certificates when talking to the '
                    'container destination registry (defaults to true)'))
@click.option('--dest-cert-dir', metavar='PATH',
              help=('use certificates at PATH (*.crt, *.cert, *.key) to connect to the '
                    'destination registry'))
@click.argument('src', metavar='SOURCE-IMAGE-NAME')
@click.argument('dest', metavar='DEST-IMAGE-NAME')
def copy(src, dest, src_creds, src_tls_verify, src_cert_dir, dest_creds, dest_tls_verify,
         dest_cert_dir):
    """Copy an image from one location to another."""
    src = parse_spec(src, src_creds, src_tls_verify, src_cert_dir)
    dest = parse_spec(dest, dest_creds, dest_tls_verify, dest_cert_dir)

    if src.registry != dest.registry:
        tempdir = tempfile.mkdtemp()
        try:
            tmp = DirectoryEndpoint(tempdir)
            Copier(src.get_endpoint(), tmp).copy()
            Copier(tmp, dest.get_endpoint()).copy()
        finally:
            shutil.rmtree(tempdir)
    else:
        Copier(src.get_endpoint(), dest.get_endpoint()).copy()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright © 2014-2018 Red Hat Inc., and others.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Defines API endpoints related to Release objects."""

import math

from cornice import Service
from cornice.validators import colander_body_validator, colander_querystring_validator
from pyramid.exceptions import HTTPNotFound
from sqlalchemy import func, distinct
from sqlalchemy.sql import or_

from bodhi.server import log, security
from bodhi.server.models import (
    Update,
    UpdateStatus,
    UpdateType,
    Build,
    BuildrootOverride,
    Package,
    Release,
    ReleaseState,
)
from bodhi.server.validators import (
    validate_tags,
    validate_enums,
    validate_updates,
    validate_packages,
    validate_release,
)
import bodhi.server.schemas
import bodhi.server.services.errors


release = Service(name='release', path='/releases/{name}',
                  description='Fedora Releases',
                  cors_origins=bodhi.server.security.cors_origins_ro)
releases = Service(name='releases', path='/releases/',
                   description='Fedora Releases',
                   factory=security.AdminACLFactory,
                   # Note, this 'rw' is not a typo.  the @comments service has
                   # a ``post`` section at the bottom.
                   cors_origins=bodhi.server.security.cors_origins_rw)


@release.get(accept="text/html", renderer="release.html",
             error_handler=bodhi.server.services.errors.html_handler)
def get_release_html(request):
    """
    Render a release given by id as HTML.

    Args:
        request (pyramid.Request): The current request.
    Returns:
        basestring: An HTML representation of the reqeusted Release.
    """
    id = request.matchdict.get('name')
    release = Release.get(id)
    if not release:
        request.errors.add('body', 'name', 'No such release')
        request.errors.status = HTTPNotFound.code
    updates = request.db.query(Update).filter(Update.release == release).order_by(
        Update.date_submitted.desc())

    updates_count = request.db.query(Update.date_submitted, Update.type).filter(
        Update.release == release).order_by(Update.date_submitted.desc())

    date_commits = {}
    dates = set()

    for update in updates_count.all():
        d = update.date_submitted
        yearmonth = str(d.year) + '/' + str(d.month).zfill(2)
        dates.add(yearmonth)
        if update.type.description not in date_commits:
            date_commits[update.type.description] = {}
        if yearmonth in date_commits[update.type.description]:
            date_commits[update.type.description][yearmonth] += 1
        else:
            date_commits[update.type.description][yearmonth] = 0

    base_count_query = request.db.query(Update)\
        .filter(Update.release == release)

    num_updates_pending = base_count_query\
        .filter(Update.status == UpdateStatus.pending).count()
    num_updates_testing = base_count_query\
        .filter(Update.status == UpdateStatus.testing).count()
    num_updates_stable = base_count_query\
        .filter(Update.status == UpdateStatus.stable).count()
    num_updates_unpushed = base_count_query\
        .filter(Update.status == UpdateStatus.unpushed).count()
    num_updates_obsolete = base_count_query\
        .filter(Update.status == UpdateStatus.obsolete).count()

    num_updates_security = base_count_query\
        .filter(Update.type == UpdateType.security).count()
    num_updates_bugfix = base_count_query\
        .filter(Update.type == UpdateType.bugfix).count()
    num_updates_enhancement = base_count_query\
        .filter(Update.type == UpdateType.enhancement).count()
    num_updates_newpackage = base_count_query\
        .filter(Update.type == UpdateType.newpackage).count()

    num_active_overrides = request.db.query(
        BuildrootOverride
    ).filter(
        BuildrootOverride.expired_date.is_(None)
    ).join(
        BuildrootOverride.build
    ).join(
        Build.release
    ).filter(
        Build.release == release
    ).count()

    num_expired_overrides = request.db.query(
        BuildrootOverride
    ).filter(
        BuildrootOverride.expired_date.isnot(None)
    ).join(
        BuildrootOverride.build
    ).join(
        Build.release
    ).filter(
        Build.release == release
    ).count()

    return dict(release=release,
                latest_updates=updates.limit(25).all(),
                count=updates.count(),
                date_commits=date_commits,
                dates=sorted(dates),

                num_updates_pending=num_updates_pending,
                num_updates_testing=num_updates_testing,
                num_updates_stable=num_updates_stable,
                num_updates_unpushed=num_updates_unpushed,
                num_updates_obsolete=num_updates_obsolete,

                num_updates_security=num_updates_security,
                num_updates_bugfix=num_updates_bugfix,
                num_updates_enhancement=num_updates_enhancement,
                num_updates_newpackage=num_updates_newpackage,

                num_active_overrides=num_active_overrides,
                num_expired_overrides=num_expired_overrides,
                )


@release.get(accept=('application/json', 'text/json'), renderer='json',
             error_handler=bodhi.server.services.errors.json_handler)
@release.get(accept=('application/javascript'), renderer='jsonp',
             error_handler=bodhi.server.services.errors.jsonp_handler)
def get_release_json(request):
    """
    Return JSON for a release given by name.

    Args:
        request (pyramid.json): The current request.
    Returns:
        bodhi.server.models.Release: The matched Release.
    """
    id = request.matchdict.get('name')
    release = Release.get(id)
    if not release:
        request.errors.add('body', 'name', 'No such release')
        request.errors.status = HTTPNotFound.code
    return release


releases_get_validators = (colander_querystring_validator, validate_release, validate_updates,
                           validate_packages)


@releases.get(accept="text/html", schema=bodhi.server.schemas.ListReleaseSchema,
              renderer='releases.html',
              error_handler=bodhi.server.services.errors.html_handler,
              validators=releases_get_validators)
def query_releases_html(request):
    """
    Return all releases, collated by state, rendered as HTML.

    Args:
        request (pyramid.request): The current request.
    Returns:
        dict: A dictionary with a single key, releases, mapping another dictionary that maps release
            states to a list of Release objects that are in that state.
    """
    def collect_releases(releases):
        x = {}
        for r in releases:
            if r['state'] in x:
                x[r['state']].append(r)
            else:
                x[r['state']] = [r]
        return x

    db = request.db
    releases = db.query(Release).order_by(Release.id.desc()).all()
    return dict(releases=collect_releases(releases))


@releases.get(accept=('application/json', 'text/json'),
              schema=bodhi.server.schemas.ListReleaseSchema, renderer='json',
              error_handler=bodhi.server.services.errors.json_handler,
              validators=releases_get_validators)
def query_releases_json(request):
    """
    Search releases by given criteria, returning the results as JSON.

    Args:
        request (pyramid.request): The current request.
    Returns:
        dict: A dictionary with the following keys:
            releases: An iterable of the Releases that match the query.
            page: The current page.
            pages: The total number of pages.
            rows_per_page: The number of rown on a page.
            total: The number of matching results.
    """
    db = request.db
    data = request.validated
    query = db.query(Release)

    ids = data.get('ids')
    if ids is not None:
        query = query.filter(or_(*[Release.id == _id for _id in ids]))

    name = data.get('name')
    if name is not None:
        query = query.filter(Release.name.like(name))

    updates = data.get('updates')
    if updates is not None:
        query = query.join(Release.builds).join(Build.update)
        args = \
            [Update.title == update.title for update in updates] +\
            [Update.alias == update.alias for update in updates]
        query = query.filter(or_(*args))

    packages = data.get('packages')
    if packages is not None:
        query = query.join(Release.builds).join(Build.package)
        query = query.filter(or_(*[Package.id == p.id for p in packages]))

    exclude_archived = data.get('exclude_archived')
    if exclude_archived:
        query = query.filter(Release.state != ReleaseState.archived)

    state = data.get('state')
    if state is not None:
        query = query.filter(Release.state == ReleaseState.from_string(state))

    # We can't use ``query.count()`` here because it is naive with respect to
    # all the joins that we're doing above.
    count_query = query.with_labels().statement\
        .with_only_columns([func.count(distinct(Release.id))])\
        .order_by(None)
    total = db.execute(count_query).scalar()

    page = data.get('page')
    rows_per_page = data.get('rows_per_page')
    pages = int(math.ceil(total / float(rows_per_page)))
    query = query.offset(rows_per_page * (page - 1)).limit(rows_per_page)

    return dict(
        releases=query.all(),
        page=page,
        pages=pages,
        rows_per_page=rows_per_page,
        total=total,
    )


@releases.post(schema=bodhi.server.schemas.SaveReleaseSchema,
               permission='admin', renderer='json',
               error_handler=bodhi.server.services.errors.json_handler,
               validators=(colander_body_validator, validate_tags, validate_enums)
               )
def save_release(request):
    """
    Save a release.

    This entails either creating a new release, or editing an existing one. To
    edit an existing release, the release's original name must be specified in
    the ``edited`` parameter.

    Args:
        request (pyramid.request): The current request.
    Returns:
        bodhi.server.models.Request: The created or edited Request.
    """
    data = request.validated

    edited = data.pop("edited", None)

    # This has already been validated at this point, but we need to ditch
    # it since the models don't care about a csrf argument.
    data.pop('csrf_token', None)

    try:
        if edited is None:
            log.info("Creating a new release: %s" % data['name'])
            r = Release(**data)

        else:
            log.info("Editing release: %s" % edited)
            r = request.db.query(Release).filter(Release.name == edited).one()
            for k, v in data.items():
                setattr(r, k, v)

    except Exception as e:
        log.exception(e)
        request.errors.add('body', 'release',
                           'Unable to create/edit release: %s' % e)
        return

    request.db.add(r)
    request.db.flush()

    return r



# Copyright (c) 2018 Red Hat, Inc.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Add a display_name field to the updates table.

Revision ID: be25565a1211
Revises: 22858ba91115
Create Date: 2018-04-12 18:54:06.477866
"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'be25565a1211'
down_revision = '22858ba91115'


def upgrade():
    """Add a display_name column to the updates table."""
    op.add_column('updates', sa.Column('display_name', sa.UnicodeText(), nullable=False,
                  server_default=u''))


def downgrade():
    """Drop the display_name field from the updates table."""
    op.drop_column('updates', 'display_name')



# -*- coding: utf-8 -*-
# Copyright © 2014-2017 Red Hat, Inc. and others
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""A collection of views that don't fit in any other common category."""

import datetime

from pyramid.settings import asbool
from pyramid.view import view_config, notfound_view_config
from pyramid.exceptions import HTTPForbidden
from pyramid.httpexceptions import HTTPFound
import cornice.errors
import sqlalchemy as sa
import six

from bodhi.server import log, models
from bodhi.server.config import config
import bodhi.server.util


def get_top_testers():
    """
    Return a query of the 5 users that have submitted the most comments in the last 7 days.

    Returns:
        sqlalchemy.orm.query.Query: A SQLAlchemy query that contains the
                                  5 users that have submitted the most comments
                                  in the last 7 days, and their total number of
                                  comments in bodhi.
    """
    blacklist = config.get('stats_blacklist')
    days = config.get('top_testers_timeframe')
    start_time = datetime.datetime.utcnow() - datetime.timedelta(days=days)

    query = models.Session().query(
        models.User,
        sa.func.count(models.User.comments).label(u'count_1')
    ).join(models.Comment)
    query = query\
        .order_by(sa.text(u'count_1 desc'))\
        .filter(models.Comment.timestamp > start_time)

    for user in blacklist:
        query = query.filter(models.User.name != six.text_type(user))

    return query\
        .group_by(models.User)\
        .limit(5)\
        .all()


def get_latest_updates(critpath, security):
    """
    Return a query of the 5 most recent updates that are in Testing status.

    critpath and security are optional filters for security or critical path updates.

    Args:
        critpath (boolean): If True, filter results to only return updates in the Critical Path
        security (boolean): If True, filter results to only return Security updates
    Returns:
        sqlalchemy.orm.query.Query: A SQLAlchemy query that contains the
                                  5 most recent updates that are in Testing status
    """
    query = models.Update.query

    if critpath:
        query = query.filter(
            models.Update.critpath == True)
    if security:
        query = query.filter(
            models.Update.type == models.UpdateType.security)

    query = query.filter(
        models.Update.status == models.UpdateStatus.testing)

    query = query.order_by(models.Update.date_submitted.desc())
    return query.limit(5).all()


def _get_status_counts(basequery, status):
    """
    Return a dictionary with the counts of objects found in the basequery.

    The return data is specified by total count, newpackage count, bugfix count, enhancement count,
    and security count. The dictionary keys will be named with the
    template {status}_{type}_total. For example, if status is
    models.UpdateStatus.stable, a dictionary with the following keys would be
    returned:
        stable_updates_total
        stable_newpackage_total
        stable_bugfix_total
        stable_enhancement_total
        stable_security_total

    Args:
        basequery (sqlalchemy.orm.query.Query):
            The basequery of updates we want to count and further filter on.
        status (bodhi.server.models.UpdateStatus):
            The update status we want to filter by in basequery
    Return:
        dict: A dictionary describing the counts of the updates, as described above.
    """
    basequery = basequery.filter(models.Update.status == status)
    return {
        '{}_updates_total'.format(status.description): basequery.count(),
        '{}_newpackage_total'.format(status.description):
            basequery.filter(models.Update.type == models.UpdateType.newpackage).count(),
        '{}_bugfix_total'.format(status.description):
            basequery.filter(models.Update.type == models.UpdateType.bugfix).count(),
        '{}_enhancement_total'.format(status.description):
            basequery.filter(models.Update.type == models.UpdateType.enhancement).count(),
        '{}_security_total'.format(status.description):
            basequery.filter(models.Update.type == models.UpdateType.security).count(),
    }


def get_update_counts(releaseid):
    """
    Return counts for the various states and types of updates in the given release.

    This function returns a dictionary that tabulates the counts of the various
    types of Bodhi updates at the various states they can appear in. The
    dictionary has the following keys, with pretty self-explanatory names:

        pending_updates_total
        pending_newpackage_total
        pending_bugfix_total
        pending_enhancement_total
        pending_security_total
        testing_updates_total
        testing_newpackage_total
        testing_bugfix_total
        testing_enhancement_total
        testing_security_total
        stable_updates_total
        stable_newpackage_total
        stable_bugfix_total
        stable_enhancement_total
        stable_security_total

    Args:
        request (pyramid.util.Request): The current request
        releaseid (basestring): The id of the Release object you would like the counts performed on
    Returns:
        dict: A dictionary expressing the counts, as described above.
    """
    release = models.Release.get(releaseid)
    basequery = models.Update.query.filter(models.Update.release == release)
    counts = {}
    counts.update(_get_status_counts(basequery, models.UpdateStatus.pending))
    counts.update(_get_status_counts(basequery, models.UpdateStatus.testing))
    counts.update(_get_status_counts(basequery, models.UpdateStatus.stable))

    return counts


def _generate_home_page_stats():
    """
    Generate and return a dictionary of stats for the home() function to use.

    This function returns a dictionary with the following 4 keys:

        release_updates_counts: a dict of counts of various states and types of updates as
                                generated by get_update_counts()
        top_testers:            a list of 5 tuples, the 5 top testers in the last 7 days.
                                The first item of each tuple contains a dict of the items
                                in the User model for that user. The second item of the tuple
                                contains the number of comments the user has left in Bodhi.
        critpath_updates:       a list of 5 dicts, the 5 most recent critical path updates.
                                Each dict contains key-value pairs for all the items in the
                                Update model.
        security_updates:       a list of 5 dicts, the 5 most recent security updates.
                                Each dict contains key-value pairs for all the items in the
                                Update model.

    Returns:
        dict: A Dictionary expressing the values described above
    """
    top_testers = get_top_testers()
    critpath_updates = get_latest_updates(True, False)
    security_updates = get_latest_updates(False, True)
    release_updates_counts = {}
    releases = models.Release.all_releases()
    for release in releases['current'] + releases['pending']:
        release_updates_counts[release["name"]] = get_update_counts(release["name"])

    return {
        "release_updates_counts": release_updates_counts,
        "top_testers": [(obj.__json__(), n) for obj, n in top_testers],
        "critpath_updates": [obj.__json__() for obj in critpath_updates],
        "security_updates": [obj.__json__() for obj in security_updates],
    }


@view_config(route_name='home', renderer='home.html')
def home(request):
    """
    Provide the data required to present the Bodhi frontpage.

    See the docblock on _generate_home_page_stats() for details on the return value.

    Args:
        request (pyramid.request): The current web request.
    Returns:
        dict: A Dictionary expressing the values described in the docblock for
            _generate_home_page_stats().
    """
    return _generate_home_page_stats()


@view_config(route_name='new_update', renderer='new_update.html')
def new_update(request):
    """
    Return the new update form.

    Args:
        request (pyramid.util.Request): The current request.
    Returns:
        dict: A dictionary with four keys. "update" indexes None. "types" indexes a list of the
            possible UpdateTypes. "severities" indexes a list of the possible severity values.
            "suggestions" indexes a list of the possible values for update suggestions.
    Raises:
        pyramid.exceptions.HTTPForbidden: If the user is not logged in.
    """
    user = request.authenticated_userid
    if not user:
        raise HTTPForbidden("You must be logged in.")
    suggestions = list(models.UpdateSuggestion.values())
    if six.PY2:  # pragma: no cover
        suggestions = reversed(suggestions)
    return dict(
        update=None,
        types=reversed(list(models.UpdateType.values())),
        severities=sorted(list(models.UpdateSeverity.values()),
                          key=bodhi.server.util.sort_severity),
        suggestions=suggestions,
    )


@view_config(route_name='latest_candidates', renderer='json')
def latest_candidates(request):
    """
    Return the most recent candidate builds for a given package name.

    For a given `package`, this method returns the most recent builds tagged
    into the Release.candidate_tag for all Releases. The package name is specified in the request
    "package" parameter.

    Args:
        request (pyramid.util.Request): The current request. The package name is specified in the
            request's "package" parameter.
    Returns:
        list: A list of dictionaries of the found builds. Each dictionary has two keys: "nvr" maps
            to the build's nvr field, and "id" maps to the build's id.
    """
    koji = request.koji
    db = request.db

    def work(pkg, testing):
        result = []
        koji.multicall = True

        releases = db.query(models.Release) \
                     .filter(
                         models.Release.state.in_(
                             (models.ReleaseState.pending,
                              models.ReleaseState.current)))

        kwargs = dict(package=pkg, latest=True)
        for release in releases:
            koji.listTagged(release.candidate_tag, **kwargs)
            if testing:
                koji.listTagged(release.testing_tag, **kwargs)
                koji.listTagged(release.pending_testing_tag, **kwargs)
                koji.listTagged(release.pending_signing_tag, **kwargs)

        response = koji.multiCall() or []  # Protect against None

        for taglist in response:
            for build in taglist[0]:
                item = {
                    'nvr': build['nvr'],
                    'id': build['id'],
                }
                # Prune duplicates
                # https://github.com/fedora-infra/bodhi/issues/450
                if item not in result:
                    result.append(item)
        return result

    pkg = request.params.get('package')
    testing = asbool(request.params.get('testing'))
    log.debug('latest_candidate(%r, %r)' % (pkg, testing))

    if not pkg:
        return []

    result = work(pkg, testing)

    log.debug(result)
    return result


@view_config(route_name='latest_builds', renderer='json')
def latest_builds(request):
    """
    Return a list of the latest builds for a given package.

    Args:
        request (pyramid.util.Request): The current request. The request's "package" parameter is
            used to pass the package name being queried.
    Returns:
        dict: A dictionary of the release dist tag to the latest build.
    """
    builds = {}
    koji = request.koji
    package = request.params.get('package')
    for tag_type, tags in six.iteritems(models.Release.get_tags(request.db)[0]):
        for tag in tags:
            try:
                for build in koji.getLatestBuilds(tag, package=package):
                    builds[tag] = build['nvr']
            except Exception:  # Things like EPEL don't have pending tags
                pass
    return builds


@view_config(route_name='masher_status', renderer='masher.html')
def masher_status(request):
    """
    Return the masher status page.

    Args:
        request (pyramid.util.Request): The current request. Unused.
    Returns:
        dict: An empty dictionary.
    """
    return dict()


@view_config(route_name='new_override', renderer='override.html')
def new_override(request):
    """
    Return the new buildroot override form.

    Args:
        request (pyramid.util.Request): The current request.
    Returns:
        dict: A dictionary of the form {nvr: nvr}, where the request nvr field indexes itself.
    Raises:
        pyramid.exceptions.HTTPForbidden: If the user is not logged in.
    """
    nvr = request.params.get('nvr')
    user = request.authenticated_userid
    if not user:
        raise HTTPForbidden("You must be logged in.")
    return dict(nvr=nvr)


@view_config(route_name='popup_toggle', request_method='POST')
def popup_toggle(request):
    """
    Toggle whether the user is configured to receive fedmsg popups in the web UI.

    Args:
        request (pyramid.util.Request): The current request.
    Returns:
        pyramid.httpexceptions.HTTPFound: A redirect to the "next" field of the request, or the home
            page if "next" is not defined.
    Raises:
        pyramid.exceptions.HTTPForbidden: If the user is not logged in.
    """
    # Get the user
    userid = request.authenticated_userid
    if userid is None:
        raise HTTPForbidden("You must be logged in.")
    user = request.db.query(models.User).filter_by(name=six.text_type(userid)).first()

    # Toggle the value.
    user.show_popups = not user.show_popups

    # And send the user back
    return_to = request.params.get('next', request.route_url('home'))
    return HTTPFound(location=return_to)


@view_config(route_name='api_version', renderer='json')
def api_version(request):
    """
    Return the Bodhi API version.

    Args:
        request (pyramid.util.Request): The current request.
    Returns:
        dict: A dictionary with a "version" key indexing a string of the Bodhi version.
    """
    return dict(version=bodhi.server.util.version())


@notfound_view_config(append_slash=True)
def notfound_view(context, request):
    """
    Automatically redirects to slash-appended routes.

    Note: The URL below spans two lines.
    http://docs.pylonsproject.org/projects/pyramid/en/latest/narr/urldispatch.html

    Args:
        context (pyramid.httpexceptions.HTTPNotFound): The 404 not found error.
        request (pyramid.util.Request): The current request.
    Returns:
        bodhi.server.services.errors.html_handler: A pyramid.httpexceptions.HTTPError to be rendered
            to the user for the 404.
    """
    return exception_view(context, request)


@view_config(context=HTTPForbidden)
@view_config(context=Exception)
def exception_view(exc, request):
    """
    Return an error response upon generic errors (404s, 403s, 500s, etc..).

    This is here to catch everything that isn't caught by our cornice error
    handlers.  When we do catch something, we transform it into a cornice
    Errors object and pass it to our nice cornice error handler.  That way, all
    the exception presentation and rendering we can keep in one place.

    Args:
        exc (Exception): The unhandled exception.
        request (pyramid.util.Request): The current request.
    Returns:
        bodhi.server.services.errors.html_handler: A pyramid.httpexceptions.HTTPError to be rendered
            to the user for the given exception.
    """
    errors = getattr(request, 'errors', [])
    status = getattr(exc, 'status_code', 500)

    if status not in (404, 403):
        log.exception("Error caught.  Handling HTML response.")
    else:
        log.warning(str(exc))

    if not len(errors):
        description = getattr(exc, 'explanation', None) or str(exc)

        errors = cornice.errors.Errors(status=status)
        errors.add('body', description=description)

    return bodhi.server.services.errors.html_handler(errors, request)



# -*- coding: utf-8 -*-
# Copyright © 2011-2017 Red Hat, Inc. and others.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Define utilities for handling errors in the service modules."""

import logging
import re

import cornice.util
import mako.exceptions
import mako.lookup
import pyramid.httpexceptions
import pyramid.response

from bodhi.server.config import config
from bodhi.server.util import get_absolute_path


log = logging.getLogger('bodhi')


# First, just re-use cornice's default json handler for our own.  It is fine.
json_handler = cornice.util.json_error

# TODO -- make this do the right thing.  Not a big deal for now.
jsonp_handler = cornice.util.json_error


def camel2space(camel):
    """
    Convert CamelCaseText to Space Separated Text.

    Args:
        camel (basestring): Camel cased text you wish to convert to space separated text.
    Returns:
        basestring: A space separated version of the given camel cased text.
    """
    regexp = r'([A-Z][a-z0-9]+|[a-z0-9]+|[A-Z0-9]+)'
    return ' '.join(re.findall(regexp, camel))


def status2summary(status):
    """
    Convert numerical HTTP status codes to human readable error strings.

    For example, this converts 404 to "Not Found".

    Args:
        status (int): The status you wish to have a human readable string for.
    Returns:
        basestring: A human readable error message.
    """
    cls = pyramid.httpexceptions.status_map[status]
    camel = cls.__name__[4:]
    return camel2space(camel)


class html_handler(pyramid.httpexceptions.HTTPError):
    """An HTML formatting handler for all our errors."""

    def __init__(self, errors, request):
        """
        Initialize the HTML error handler to render an error messgae for human readers.

        This method sets the Response body to rendered HTML version of the given errors, and the
        status code to the code specified by errors.

        Args:
            errors (cornice.errors.Errors): The errors to be rendered as HTML for users.
            request (pyramid.util.Request): The current Request.
        """
        location = config.get('mako.directories')
        directory = get_absolute_path(location)

        lookup = mako.lookup.TemplateLookup(
            directories=[directory],
            output_encoding='utf-8',
            input_encoding='utf-8',
        )
        template = lookup.get_template('errors.html')

        try:
            body = template.render(
                errors=errors,
                status=errors.status,
                request=request,
                summary=status2summary(errors.status),
            )
        except Exception:
            log.error(mako.exceptions.text_error_template().render())
            raise

        # This thing inherits from both Exception *and* Response, so.. take the
        # Response path in the diamond inheritance chain and ignore the
        # exception side.
        # That turns this thing into a "real boy" like pinnochio.
        pyramid.response.Response.__init__(self, body)

        self.status = errors.status
        self.content_type = 'text/html'



"""Drop the unused inherited column from the builds table.

Revision ID: 12d3e8695f90
Revises: fc6b0169c596
Create Date: 2017-04-12 23:26:25.293009
"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '12d3e8695f90'
down_revision = 'fc6b0169c596'


def upgrade():
    """Remove the unused inherited column from the builds table."""
    op.drop_column('builds', 'inherited')


def downgrade():
    """Add the inherited column back and set all records to False."""
    op.add_column('builds',
                  sa.Column('inherited', sa.BOOLEAN(), autoincrement=False, nullable=True))
    # Build a fake mini version of the builds table so we can form an UPDATE statement.
    builds = sa.sql.table('builds', sa.sql.column('inherited', sa.Boolean))
    # Set all records to False, since that was the default before and since there was no way to set
    # it to anything else through the API.
    op.execute(builds.update().values({'inherited': False}))



# Copyright (c) 2017 Red Hat, Inc.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
build.package_id is now required.

Revision ID: 865d9432fa7d
Revises: 74cfbc6116ad
Create Date: 2017-12-07 22:17:59.312908
"""
from alembic import op
import sqlalchemy as sa


revision = '865d9432fa7d'
down_revision = '74cfbc6116ad'


def upgrade():
    """Mark builds.package_id as non-nullable."""
    op.alter_column('builds', 'package_id', existing_type=sa.INTEGER(), nullable=False)


def downgrade():
    """Mark builds.package_id as nullable."""
    op.alter_column('builds', 'package_id', existing_type=sa.INTEGER(), nullable=True)



# -*- coding: utf-8 -*-
# Copyright © 2017 Till Maas
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Add indexes for status and request.

Revision ID: 2616c86d8ac6
Revises: da710ff02641
Create Date: 2017-10-26 20:45:23.152139
"""
from alembic import op


# revision identifiers, used by Alembic.
revision = '2616c86d8ac6'
down_revision = 'da710ff02641'


def upgrade():
    """Add indices to the request and status columns on the update table."""
    op.create_index(op.f('ix_updates_request'), 'updates', ['request'], unique=False)
    op.create_index(op.f('ix_updates_status'), 'updates', ['status'], unique=False)


def downgrade():
    """Drop indices to the request and status columns on the update table."""
    op.drop_index(op.f('ix_updates_status'), table_name='updates')
    op.drop_index(op.f('ix_updates_request'), table_name='updates')



# -*- coding: utf-8 -*-
# Copyright 2007-2018 Red Hat, Inc. and others.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Create metadata files when mashing repositories."""
from datetime import datetime
import logging
import os
import shelve
import shutil
import tempfile

from kitchen.text.converters import to_bytes
import createrepo_c as cr
import six

from bodhi.server import util
from bodhi.server.buildsys import get_session
from bodhi.server.config import config
from bodhi.server.models import Build, UpdateStatus, UpdateRequest, UpdateSuggestion


__version__ = '2.0'
log = logging.getLogger(__name__)


def insert_in_repo(comp_type, repodata, filetype, extension, source):
    """
    Inject a file into the repodata with the help of createrepo_c.

    Args:
        comp_type (int): createrepo_c compression type indication.
        repodata (basestring): The path to the repo where the metadata will be inserted.
        filetype (basestring): What type of metadata will be inserted by createrepo_c.
            This does allow any string to be inserted (custom types). There are some
            types which are used with dnf repos as primary, updateinfo, comps, filelist etc.
        extension (basestring): The file extension (xml, sqlite).
        source (basestring): A file path. File holds the dump of metadata until
            copied to the repodata folder.
    """
    log.info('Inserting %s.%s into %s', filetype, extension, repodata)
    target_fname = os.path.join(repodata, '%s.%s' % (filetype, extension))
    shutil.copyfile(source, target_fname)
    repomd_xml = os.path.join(repodata, 'repomd.xml')
    repomd = cr.Repomd(repomd_xml)
    # create a new record for our repomd.xml
    rec = cr.RepomdRecord(filetype, target_fname)
    # compress our metadata file with the comp_type
    rec_comp = rec.compress_and_fill(cr.SHA256, comp_type)
    # add hash to the compresed metadata file
    rec_comp.rename_file()
    # set type of metadata
    rec_comp.type = filetype
    # insert metadata about our metadata in repomd.xml
    repomd.set_record(rec_comp)
    with open(repomd_xml, 'w') as repomd_file:
        repomd_file.write(repomd.xml_dump())
    os.unlink(target_fname)


def modifyrepo(comp_type, compose_path, filetype, extension, source):
    """
    Inject a file into the repodata for each architecture with the help of createrepo_c.

    Args:
        compose_path (basestring): The path to the compose where the metadata will be inserted.
        filetype (basestring): What type of metadata will be inserted by createrepo_c.
            This does allow any string to be inserted (custom types). There are some
            types which are used with dnf repos as primary, updateinfo, comps, filelist etc.
        extension (basestring): The file extension (xml, sqlite).
        source (basestring): A file path. File holds the dump of metadata until
            copied to the repodata folder.
    """
    repo_path = os.path.join(compose_path, 'compose', 'Everything')
    for arch in os.listdir(repo_path):
        if arch == 'source':
            repodata = os.path.join(repo_path, arch, 'tree', 'repodata')
        else:
            repodata = os.path.join(repo_path, arch, 'os', 'repodata')
        insert_in_repo(comp_type, repodata, filetype, extension, source)


class UpdateInfoMetadata(object):
    """
    This class represents the updateinfo.xml yum metadata.

    It is generated during push time by the bodhi masher based on koji tags
    and is injected into the yum repodata using the `modifyrepo_c` tool,
    which is included in the `createrepo_c` package.
    """

    def __init__(self, release, request, db, mashdir, close_shelf=True):
        """
        Initialize the UpdateInfoMetadata object.

        Args:
            release (bodhi.server.models.Release): The Release that is being mashed.
            request (bodhi.server.models.UpdateRequest): The Request that is being mashed.
            db (): A database session to be used for queries.
            mashdir (basestring): A path to the mashdir.
            close_shelf (bool): Whether to close the shelve, which is used to cache updateinfo
                between mashes.
        """
        self.request = request
        if request is UpdateRequest.stable:
            self.tag = release.stable_tag
        else:
            self.tag = release.testing_tag

        self.db = db
        self.updates = set()
        self.builds = {}
        self._from = config.get('bodhi_email')
        if config.get('cache_dir'):
            self.shelf = shelve.open(os.path.join(config.get('cache_dir'), '%s.shelve' % self.tag))
        else:
            # If we have no cache dir, let's at least cache in-memory.
            self.shelf = {}
            close_shelf = False
        self._fetch_updates()

        self.uinfo = cr.UpdateInfo()

        self.comp_type = cr.XZ

        if release.id_prefix == u'FEDORA-EPEL':
            # FIXME: I'm not sure which versions of RHEL support xz metadata
            # compression, so use the lowest common denominator for now.
            self.comp_type = cr.BZ2

        self.uinfo = cr.UpdateInfo()
        for update in self.updates:
            if not update.alias:
                update.assign_alias()
            self.add_update(update)

        if close_shelf:
            self.shelf.close()

    def _fetch_updates(self):
        """Based on our given koji tag, populate a list of Update objects."""
        log.debug("Fetching builds tagged with '%s'" % self.tag)
        kojiBuilds = get_session().listTagged(self.tag, latest=True)
        nonexistent = []
        log.debug("%d builds found" % len(kojiBuilds))
        for build in kojiBuilds:
            self.builds[build['nvr']] = build
            build_obj = self.db.query(Build).filter_by(nvr=six.text_type(build['nvr'])).first()
            if build_obj:
                if build_obj.update:
                    self.updates.add(build_obj.update)
                else:
                    log.warning('%s does not have a corresponding update' % build['nvr'])
            else:
                nonexistent.append(build['nvr'])
        if nonexistent:
            log.warning("Couldn't find the following koji builds tagged as "
                        "%s in bodhi: %s" % (self.tag, nonexistent))

    def get_rpms(self, koji, nvr):
        """
        Retrieve the given RPM nvr from the cache if available, or from Koji if not available.

        Args:
            koji (koji.ClientSession): An initialized Koji client.
            nvr (basestring): The nvr for which you wish to retrieve Koji data.
        Returns:
            list: A list of dictionaries describing all the subpackages that are part of the given
                nvr.
        """
        if str(nvr) in self.shelf:
            return self.shelf[str(nvr)]

        if nvr in self.builds:
            buildid = self.builds[nvr]['id']
        else:
            buildid = koji.getBuild(nvr)['id']

        rpms = koji.listBuildRPMs(buildid)
        self.shelf[str(nvr)] = rpms
        return rpms

    def add_update(self, update):
        """
        Generate the extended metadata for a given update, adding it to self.uinfo.

        Args:
            update (bodhi.server.models.Update): The Update to be added to self.uinfo.
        """
        rec = cr.UpdateRecord()
        rec.version = __version__
        rec.fromstr = config.get('bodhi_email')
        rec.status = update.status.value
        rec.type = update.type.value
        rec.id = to_bytes(update.alias)
        rec.title = to_bytes(update.title)
        rec.severity = util.severity_updateinfo_str(update.severity.value)
        rec.summary = to_bytes('%s %s update' % (update.get_title(),
                                                 update.type.value))
        rec.description = to_bytes(update.notes)
        rec.release = to_bytes(update.release.long_name)
        rec.rights = config.get('updateinfo_rights')

        if update.date_pushed:
            rec.issued_date = update.date_pushed
        else:
            # Sometimes we only set the date_pushed after it's pushed out, however,
            # it seems that Satellite does not like update entries without issued_date.
            # Since we know that we are pushing it now, and the next push will get the data
            # correctly, let's just insert utcnow().
            rec.issued_date = datetime.utcnow()
        if update.date_modified:
            rec.updated_date = update.date_modified
        else:
            rec.updated_date = datetime.utcnow()

        col = cr.UpdateCollection()
        col.name = to_bytes(update.release.long_name)
        col.shortname = to_bytes(update.release.name)

        koji = get_session()
        for build in update.builds:
            rpms = self.get_rpms(koji, build.nvr)
            for rpm in rpms:
                pkg = cr.UpdateCollectionPackage()
                pkg.name = rpm['name']
                pkg.version = rpm['version']
                pkg.release = rpm['release']
                if rpm['epoch'] is not None:
                    pkg.epoch = str(rpm['epoch'])
                else:
                    pkg.epoch = '0'
                pkg.arch = rpm['arch']

                # TODO: how do we handle UpdateSuggestion.logout, etc?
                pkg.reboot_suggested = update.suggest is UpdateSuggestion.reboot

                filename = '%s.%s.rpm' % (rpm['nvr'], rpm['arch'])
                pkg.filename = filename

                # Build the URL
                if rpm['arch'] == 'src':
                    arch = 'SRPMS'
                elif rpm['arch'] in ('noarch', 'i686'):
                    arch = 'i386'
                else:
                    arch = rpm['arch']

                pkg.src = os.path.join(
                    config.get('file_url'),
                    update.status is UpdateStatus.testing and 'testing' or '',
                    str(update.release.version), arch, filename[0], filename)

                col.append(pkg)

        rec.append_collection(col)

        # Create references for each bug
        for bug in update.bugs:
            ref = cr.UpdateReference()
            ref.type = 'bugzilla'
            ref.id = to_bytes(bug.bug_id)
            ref.href = to_bytes(bug.url)
            ref.title = to_bytes(bug.title)
            rec.append_reference(ref)

        # Create references for each CVE
        for cve in update.cves:
            ref = cr.UpdateReference()
            ref.type = 'cve'
            ref.id = to_bytes(cve.cve_id)
            ref.href = to_bytes(cve.url)
            rec.append_reference(ref)

        self.uinfo.append(rec)

    def insert_updateinfo(self, compose_path):
        """
        Add the updateinfo.xml file to the repository.

        Args:
            compose_path (basestring): The path to the compose where the metadata will be inserted.
        """
        fd, tmp_file_path = tempfile.mkstemp()
        os.write(fd, self.uinfo.xml_dump().encode('utf-8'))
        os.close(fd)
        modifyrepo(self.comp_type, compose_path, 'updateinfo', 'xml', tmp_file_path)
        os.unlink(tmp_file_path)



# -*- coding: utf-8 -*-
# Copyright © 2015-2018 Red Hat, Inc. and others.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Used to remove the pending and testing tags from updates in a branched release.

Since a separate task mashes the branched stable repos, this will leave
those stable updates with the testing tags for 1 day before untagging.

https://github.com/fedora-infra/bodhi/issues/576
"""

import os
import sys
import logging

from datetime import datetime, timedelta

from pyramid.paster import get_appsettings, setup_logging

from ..models import Release, ReleaseState, Update, UpdateStatus

from bodhi.server import buildsys, Session, initialize_db


def usage(argv):
    """
    Print usage information and exit with code 1.

    Args:
        argv (list): The arguments passed to the script.
    """
    cmd = os.path.basename(argv[0])
    print('usage: %s <config_uri>\n'
          '(example: "%s development.ini")' % (cmd, cmd))
    sys.exit(1)


def main(argv=sys.argv):
    """
    Remove the pending and testing tags from branched updates.

    Args:
        argv (list): The arguments passed to the script. Defaults to sys.argv.
    """
    if len(argv) != 2:
        usage(argv)

    config_uri = argv[1]

    setup_logging(config_uri)
    log = logging.getLogger(__name__)

    settings = get_appsettings(config_uri)
    initialize_db(settings)
    db = Session()
    koji = buildsys.get_session()
    one_day = timedelta(days=1)
    now = datetime.utcnow()

    try:
        for release in db.query(Release).filter_by(
                state=ReleaseState.pending).all():
            log.info(release.name)
            for update in db.query(Update).filter_by(
                    release=release, status=UpdateStatus.stable).all():
                assert update.date_stable, update.title
                if now - update.date_stable > one_day:
                    for build in update.builds:
                        tags = build.get_tags()
                        stable_tag = release.dist_tag
                        testing_tag = release.testing_tag
                        pending_signing_tag = release.pending_signing_tag
                        pending_testing_tag = release.pending_testing_tag
                        if stable_tag not in tags:
                            log.error('%s not tagged as stable %s' % (build.nvr, tags))
                            continue
                        if testing_tag in tags:
                            log.info('Removing %s from %s' % (testing_tag, build.nvr))
                            koji.untagBuild(testing_tag, build.nvr)
                        if pending_signing_tag in tags:
                            log.info('Removing %s from %s' % (pending_signing_tag, build.nvr))
                            koji.untagBuild(pending_signing_tag, build.nvr)
                        if pending_testing_tag in tags:
                            log.info('Removing %s from %s' % (pending_testing_tag, build.nvr))
                            koji.untagBuild(pending_testing_tag, build.nvr)
        db.commit()
    except Exception as e:
        log.error(e)
        db.rollback()
        Session.remove()
        sys.exit(1)



# -*- coding: utf-8 -*-
# Copyright © 2014-2017 Red Hat, Inc. and others.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Define the view that presents release metrics in the web UI."""

import json

from pyramid.view import view_config

import bodhi.server.models as m


def compute_ticks_and_data(db, releases, update_types):
    """
    Return the data and ticks to make the stats graph.

    Args:
        db (sqlalchemy.orm.session.Session): The database Session.
        releases (list): A list of release objects we are interested in generating metrics on.
        update_types (dict): A dictionary mapping the possible types of updates to human readable
            string defining them.
    Returns:
        tuple: A 2-tuple of data that can be graphed by the UI javascript.
    """
    data, ticks = [], []

    releases = sorted(releases, key=lambda x: int(x.version_int))

    for i, release in enumerate(releases):
        ticks.append([i, release.name])

    for update_type, label in update_types.items():
        d = []
        update_type = m.UpdateType.from_string(update_type)
        for i, release in enumerate(releases):
            num = db.query(m.Update).filter_by(
                release=release,
                type=update_type,
                status=m.UpdateStatus.stable
            ).count()
            d.append([i, num])
        data.append(dict(data=d, label=label))

    return (data, ticks)


@view_config(route_name='metrics', renderer='metrics.html')
def metrics(request):
    """
    Return a response with metric data to be graphed.

    Args:
        request (pyramid.util.Request): The current Request.
    Returns:
        dict: A dictionary with keys 'data', 'ticks', 'eldata', and 'elticket'. The 'el' prefixed
            keys are for enterprise Linux. These data are used to render the graphs by the template
            JavaScript.
    """
    db = request.db

    update_types = {
        'bugfix': 'Bug fixes',
        'enhancement': 'Enhancements',
        'security': 'Security updates',
        'newpackage': 'New packages'
    }

    releases = db.query(m.Release).filter(m.Release.name.like(u'F%')).all()
    data, ticks = compute_ticks_and_data(db, releases, update_types)

    releases = db.query(m.Release).filter(m.Release.name.like(u'E%')).all()
    eldata, elticks = compute_ticks_and_data(db, releases, update_types)

    return {
        'data': json.dumps(data), 'ticks': json.dumps(ticks),
        'eldata': json.dumps(eldata), 'elticks': json.dumps(elticks),
    }



# -*- coding: utf-8 -*-
# Copyright © 2017 Red Hat, Inc.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Add ci_status to builds.

Revision ID: 4b357c65441e
Revises: b01a62d98aa4
Create Date: 2017-05-11 20:13:41.879435
"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '4b357c65441e'
down_revision = 'b01a62d98aa4'


def upgrade():
    """Add the ci_status to builds with the corresponding enum."""
    op.execute(
        "CREATE TYPE ck_ci_status AS ENUM "
        "('ignored', 'queued', 'running', 'passed', 'failed', 'waiting')")

    op.add_column(
        'builds',
        sa.Column(
            'ci_status',
            sa.Enum(
                'ignored', 'queued', 'running', 'passed', 'failed', 'waiting',
                name='ck_ci_status'),
            nullable=True
        )
    )


def downgrade():
    """Remove the ci_status from builds with the corresponding enum."""
    op.drop_column('builds', 'ci_status')
    op.execute("DROP TYPE ck_ci_status")



# -*- coding: utf-8 -*-
# Copyright © 2017 Caleigh Runge-Hottman and Red Hat, Inc.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""This script is responsible for moving all updates with a batched request to a stable request."""

import sys

import click

from bodhi.server import buildsys, config, models, Session, initialize_db


@click.command()
@click.version_option(message='%(version)s')
def dequeue_stable():
    """Convert all batched requests to stable requests."""
    initialize_db(config.config)
    buildsys.setup_buildsystem(config.config)
    db = Session()

    try:
        batched = db.query(models.Update).filter_by(request=models.UpdateRequest.batched).all()
        for update in batched:
            try:
                update.set_request(db, models.UpdateRequest.stable, u'bodhi')
                db.commit()
            except Exception as e:
                print('Unable to stabilize {}: {}'.format(update.alias, str(e)))
                db.rollback()
                msg = u"Bodhi is unable to request this update for stabilization: {}"
                update.comment(db, msg.format(str(e)), author=u'bodhi')
                db.commit()
    except Exception as e:
        print(str(e))
        sys.exit(1)
    finally:
        Session.remove()



# This should get merged smartly with the site-wide fedmsg.d/logging.py we have
# installed everywhere.
config = dict(
    logging=dict(
        loggers=dict(
            bodhi={
                "level": "DEBUG",
                "propagate": False,
                "handlers": ["console"],
            },
            sqlalchemy={
                "level": "WARN",
                "propagate": False,
                "handlers": ["console"],
            },
            root={
                "level": "INFO",
                "propagate": False,
                "handlers": ["console"],
            },
        ),
    ),
)



# -*- coding: utf-8 -*-
# Copyright © 2014-2017 Red Hat, Inc. and others.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Define an API endpoint to process text through Bodhi's markdown system."""

from cornice import Service

import bodhi.server.util
import bodhi.server.security
import bodhi.server.services.errors


markdown = Service(name='markdowner', path='/markdown',
                   description='Markdown service',
                   cors_origins=bodhi.server.security.cors_origins_ro)


@markdown.get(accept=('application/json', 'text/json'), renderer='json',
              error_handler=bodhi.server.services.errors.json_handler)
@markdown.get(accept=('application/javascript'), renderer='jsonp',
              error_handler=bodhi.server.services.errors.jsonp_handler)
def markdowner(request):
    """
    Given some text, return the markdownified html version.

    We use this for "previews" of comments and update notes.

    Args:
        request (pyramid.request): The current request. It's "text" parameter is used to specify the
            text to be processed.
    Returns:
        basestring: A JSON object with a single key, html, that indexes the processed text.
    """
    text = request.params.get('text')
    return dict(html=bodhi.server.util.markup(request.context, text))
